import numpy as np
import pytest

import pynbody


def test_indexing():
    f1 = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")

    np.random.seed(1)
    for test_len in [100, 10000, 20000]:
        for i in range(5):
            subindex = np.random.permutation(np.arange(0, len(f1)))[:test_len]
            subindex.sort()
            f2 = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024", take=subindex)

            assert (f2['x'] == f1[subindex]['x']).all()
            assert (f2['iord'] == f1[subindex]['iord']).all()

def test_load_copy():
    f1 = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")

    subview = f1[::5]

    f_subview = subview.load_copy()

    assert (subview['x']==f_subview['x']).all()

    # sanity check that the loaded copy is not linked to the original:
    subview['x'][0]=0
    f_subview['x'][0]=1
    assert subview['x'][0]==0

    with pytest.raises(NotImplementedError):
        f_subview[:5].load_copy()

def test_grp_load_copy():
    f1 = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")
    h = f1.halos()
    h1_copy = h[1].load_copy()
    assert (h1_copy['x']==h[1]['x']).all()
    assert h1_copy.ancestor is h1_copy



"""Cython code to find the next stop in a sorted list of integers."""

import numpy as np

cimport numpy as np


def scan_for_next_stop(np.ndarray[np.int64_t, ndim=1, mode="c"] ids not None,
                       np.int64_t offset_start, np.int64_t id_maximum) :
    """Scan for the next stop in a sorted list of integers.

    Parameters
    ----------

    ids : np.ndarray[np.int64_t, ndim=1, mode="c"]
        A sorted list of integers.
    offset_start : int
        The index in the list to start the search from.
    id_maximum : int
        The maximum value to search for.

    Returns
    -------

    index : int
        The index into the list of the last element that is less than or equal to *id_maximum*, or -1 if the search
        failed.

    """
    if len(ids)==0 :
            return 0
    if ids[-1]<=id_maximum :
        return len(ids)
    if ids[0]>id_maximum :
        return 0

    cdef np.int64_t left, right, mid, itr
    left,right,mid,itr = 0,0,0,0

    left = offset_start
    right = len(ids)-1
    mid = (left+right)//2

    while ids[mid-1]>id_maximum or ids[mid]<=id_maximum :
        if ids[mid]<=id_maximum :
            left = mid
        else :
            right = mid-1
        mid = (left+right+1)//2
        itr+=1
        if itr>1000 :
            break

    if itr>1000 :
        return -1
    else :
        return mid



Jtot = np.sqrt(((np.multiply(h[i]['j'].transpose(),h[i]['mass']).sum(axis=1))**2).sum())            # calculate angular momentum
W = np.sum(h[i]['phi']*h[i]['mass']) # halo potential energy
K = np.sum(h[i]['ke']*h[i]['mass'])  # halo kinetic energy
absE = np.fabs(W+K)    # total halo energy
mvir=np.sum(h[i]['mass'].in_units('Msol'))  # virial mass
rvir=pynbody.array.SimArray(np.max(h[i]['r'].in_units('kpc')),'kpc')
# 3D density profile
rhoprof = profile.Profile(h[i],dim=3,type='log')
# Rotation curve
rcpro = profile.Profile(h[i], type='equaln', nbins = 50, rmax = '40 kpc')
# surface brightness profile
diskstars = h[i].star[filt.Disc('20 kpc','3 kpc')]
sbprof = profile.Profile(diskstars,type='equaln')
# Kinematic decomposition
#decompprof = pynbody.analysis.decomp(h[i])
#dec = h[i].star['decomp']

### Save important numbers using pickle.
pickle.dump({'z':s.properties['z'],
             'time':s.properties['time'].in_units('Gyr'),
             'rvir':rvir,
             'mvir':mvir,
             'mgas': np.sum(h[i].gas['mass'].in_units('Msol')),
             'mstar': np.sum(h[i].star['mass'].in_units('Msol')),
#             'mdisk': np.sum(h[i].star[np.where(dec == 1)]['mass'].in_units('Msol')),
#             'msphere': np.sum(h[i].star[np.where(dec == 2)]['mass'].in_units('Msol')),
#             'mbulge': np.sum(h[i].star[np.where(dec == 3)]['mass'].in_units('Msol')),
#             'mthick': np.sum(h[i].star[np.where(dec == 4)]['mass'].in_units('Msol')),
#             'mpseudob': np.sum(h[i].star[np.where(dec == 5)]['mass'].in_units('Msol')),
             'mgashot': np.sum(h[i].gas[filt.HighPass('temp',1e5)]['mass'].in_units('Msol')),
             'mgascool': np.sum(h[i].gas[filt.LowPass('temp',1e5)]['mass'].in_units('Msol')),
             'Jtot':Jtot,'lambda':(Jtot / np.sqrt(2*np.power(mvir,3)*rvir*units.G)).in_units('1'),
             'denprof':{'r':rhoprof['rbins'].in_units('kpc'),
                        'den':rhoprof['density']},
             'rotcur':{'r':rcpro['rbins'].in_units('kpc'),
                       'vc':rcpro['rotation_curve_spherical'].in_units('km s^-1'),
                       'fourier':rcpro['fourier']},
             'sb':{'r':sbprof['rbins'].in_units('kpc'),
                   'sb':sbprof['sb,i']}
             },
            open(simname+'.data','w'), pickle.HIGHEST_PROTOCOL)



"""

Defines the :class:`Family` class which represents families of particles
(e.g. dm, gas, star particles).

Most users will not need to use this module directly. New families of particles
can, however, be added interactively by creating new instances of the
:class:`Family` class. See the example in the class documentation.

Another way to add families, more permanently, is to use the configuration file.
See the :ref:`configuration` section for more information.
"""

import functools

from . import config_parser

_registry = []


def family_names(with_aliases=False):
    """Returns a list of the names of all particle families.

    Parameters
    ----------

    with_aliases : bool
        If True, the list will include all aliases for each family. Default is False.

    Returns
    -------

    list of str
        The list of family names.

    """

    global _registry
    l = []
    for o in _registry:
        l.append(o.name)
        if with_aliases:
            for a in o.aliases:
                l.append(a)
    return l


def get_family(name, create=False):
    """Returns a :class:`Family` object corresponding to the specified string.

    Parameters
    ----------

    name : str or :class:`Family`
        The name of the family to retrieve. If a :class:`Family` object is passed, it is returned unchanged.

    create : bool
        If True, a new family is created if the specified family does not exist. Default is False.
        If False, a ``ValueError`` is raised if the specified family does not exist.

    """

    if isinstance(name, Family):
        return name

    name = name.lower()
    # or should it check and raise rather than just convert?
    # Not sure.
    for n in _registry:
        if n.name == name or name in n.aliases:
            return n

    if create:
        return Family(name)
    else:
        raise ValueError(name + " is not a family")

@functools.total_ordering
class Family:
    """The class representing an abstract family of particles.

    Once a :class:`Family` is defined, it is automatically registered and
    can be used within :class:`~pynbody.snapshot.simsnap.SimSnap` objects. For example,

    >>> import pynbody
    >>> tachs = pynbody.family.Family("tachyon", aliases=["t", "tachyons"])
    >>> f = pynbody.new(dm=100, tachyon=100)
    >>> f.tachyon['pos'] # <-- get the tachyon positions
    >>> f.t['pos'] # <-- use an alias to get the tachyon positions
    >>> f[tachs]['pos'] # <-- another way to get the tachyon positions

    """

    def __init__(self, name, aliases=[]):
        """Create a new family with the specified name and aliases.

        Parameters
        ----------

        name : str
            The name of the family. Must be lower case and unique.

        aliases : list of str
            A list of aliases for the family. Must be lower case. These can be used
            interchangeably with the family name.
        """
        if name != name.lower():
            raise ValueError("Family names must be lower case")
        if name in family_names(with_aliases=True):
            raise ValueError("Family name " + name + " is not unique")
        for a in aliases:
            if a != a.lower():
                raise ValueError("Aliases must be lower case")

        self.name = name
        self.aliases = aliases
        _registry.append(self)

    def __repr__(self):
        return "<Family " + self.name + ">"

    def __reduce__(self):
        return get_family, (self.name, True), {"aliases": self.aliases}

    def __iter__(self):
        # Provided so a single family can be treated as a list of families
        yield self

    def __str__(self):
        return self.name

    def __cmp__(self, other):
        # for python 2.x
        return cmp(str(self), str(other))

    def __eq__(self, other):
        return str(self) == str(other)

    def __lt__(self, other):
        return str(self) < str(other)

    def __hash__(self):
        return hash(str(self))

# Instantiate the default families as specified by the configuration file

g = globals()
for f in config_parser.options('families'):
    aliases = config_parser.get('families', f)
    g[f] = Family(f, list(map(str.strip, aliases.split(","))))



"""
Implements classes to load and manipulate snapshot data
"""

import logging
import pathlib

from .. import config, family
from . import util
from .simsnap import SimSnap

logger = logging.getLogger('pynbody.snapshot')


def load(filename, *args, **kwargs) -> SimSnap:
    """Loads a file using the appropriate class, returning a SimSnap instance.

    This routine is the main entry point for loading snapshots. It will try to load the file using the appropriate
    class, based on inspection by the candidate subclasses. If no class can load the file, an OSError is raised.

    Arguments
    ---------
    filename : str
        The filename to load

    priority : optional, list[str | type]
        A list of SimSnap subclasses to try, in order. The first class which is capable of loading the file
        is used. If not specified, the ordering is as specified in the configuration files.

    *args, **kwargs :
        Other arguments and keyword arguments are passed to the class constructor that is used to load the file.

    Returns
    -------
    SimSnap
        The loaded snapshot

    """

    filename = pathlib.Path(filename)

    priority = kwargs.pop('priority', config['snap-class-priority'])

    for c in SimSnap.iter_subclasses_with_priority(priority):
        if c._can_load(filename):
            logger.info("Loading using backend %s" % str(c))
            return c(filename, *args, **kwargs)

    raise OSError(
        "File %r: format not understood or does not exist" % filename)

def new(n_particles = 0, order = None, class_ = SimSnap, **families) -> SimSnap:
    """Create a blank SimSnap, with the specified number of particles.

    Position, velocity and mass arrays are created and filled with zeros.

    By default all particles are taken to be dark matter.

    To specify otherwise, pass in keyword arguments specifying the number of particles for each family, e.g.

    >>> f = new(dm=50, star=25, gas=25)

    The order in which the different families appear in the snapshot is unspecified unless you add an 'order' argument:

    >>> f = new(dm=50, star=25, gas=25, order='star,gas,dm')

    guarantees the stars, then gas, then dark matter particles appear in sequence.
    """

    if len(families) == 0:
        families = {'dm': n_particles}

    t_fam = []
    tot_particles = 0

    if order is None:
        for k, v in list(families.items()):

            assert isinstance(v, int)
            t_fam.append((family.get_family(k), v))
            tot_particles += v
    else:
        for k in order.split(","):
            v = families[k]
            assert isinstance(v, int)
            t_fam.append((family.get_family(k), v))
            tot_particles += v

    x = class_()
    x._num_particles = tot_particles
    x._filename = "<created>"

    x._create_arrays(["pos", "vel"], 3)
    x._create_arrays(["mass"], 1)

    rt = 0
    for k, v in t_fam:
        x._family_slice[k] = slice(rt, rt + v)
        rt += v

    x._decorate()
    return x


from . import ascii, gadget, gadgethdf, grafic, nchilada, ramses, subsnap, swift, tipsy
from .subsnap import FamilySubSnap, IndexedSubSnap, SubSnap



"""

gravity
=======

Functions for calculating the gravitational potential and accelerations.

"""


import math

import numpy as np

from .. import array, units


def potential(f, pos_vec, eps=None, unit=None):
    """

    Calculates the gravitational potential at the specified position
    using all particles in the specified snapshot.

    The gravitational softening length is determined by (in order of
    preference):

    1. The parameter eps (scalar, unit or array)

    2. The array f['eps']

    3. f.properties['eps'] (scalar or unit)

    """

    if eps is None:
        try:
            eps = f['eps']
        except KeyError:
            eps = f.properties['eps']

    if isinstance(eps, str):
        eps = units.Unit(eps)

    if isinstance(eps, units.UnitBase):
        eps = eps.in_units(f['pos'].units, **f.conversion_context())

    d_pos = f['pos'] - pos_vec

    GM_by_r = units.G * f['mass'] / \
        ((d_pos ** 2).sum(axis=1) + eps ** 2) ** (1, 2)

    if unit is not None:
        GM_by_r.convert_units(unit)

    return -(GM_by_r.sum())


def accel(f, pos_vec, eps=None):
    """

    Calculates the gravitational acceleration vector at the specified
    position using all particles in the specified snapshot.

    The gravitational softening length is determined by (in order of
    preference):

    1. The parameter eps (scalar, unit or array)

    2. The array f['eps']

    3. f.properties['eps'] (scalar or unit)

    """

    if eps is None:
        try:
            eps = f['eps']
        except KeyError:
            eps = f.properties['eps']

    if isinstance(eps, str):
        eps = units.Unit(eps)

    if isinstance(eps, units.UnitBase):
        eps = eps.in_units(f['pos'].units, **f.conversion_context())

    d_pos = f['pos'] - pos_vec

    GM_by_r3 = units.G * f['mass'] / \
        ((d_pos ** 2).sum(axis=1) + eps ** 2) ** (3, 2)

    return -(d_pos.T * GM_by_r3).sum(axis=1)


def midplane_rot_curve(f, rxy_points, eps=None):
    """

    Calculates the midplane rotation curve at a specified set of
    points. Used primarily by :func:`pynbody.analysis.profile.v_circ`

    """

    u_out = (units.G * f['mass'].units / f['pos'].units) ** (1, 2)

    vels = []

    for r in rxy_points:
        # Do four samples like Tipsy does
        r_acc_r = []
        for pos in [(r, 0, 0), (0, r, 0), (-r, 0, 0), (0, -r, 0)]:
            acc = accel(f, pos, eps)
            r_acc_r.append(np.dot(acc, pos))

        vel2 = np.mean(r_acc_r)
        if vel2 > 0:
            vel = math.sqrt(vel2)
        else:
            vel = 0

        vels.append(vel)

    x = array.SimArray(vels, units=u_out)
    x.sim = f.ancestor
    return x


def midplane_potential(f, rxy_points, eps=None):
    """

    Calculates the midplane potential at a specified set of
    points. Used primarily by :func:`pynbody.analysis.profile.pot`

    """

    u_out = units.G * f['mass'].units / f['pos'].units

    pots = []

    for r in rxy_points:
        # Do four samples like Tipsy does
        pot = []
        for pos in [(r, 0, 0), (0, r, 0), (-r, 0, 0), (0, -r, 0)]:
            pot.append(potential(f, pos, eps))

        pots.append(np.mean(pot))

    x = array.SimArray(pots, units=u_out)
    x.sim = f.ancestor
    return x



"""Gravity calculations
"""

from __future__ import annotations

import math
import warnings

import numpy as np

from .. import array, config, units
from ..array import SimArray
from ..snapshot.simsnap import SimSnap
from ..util import eps_as_simarray, get_eps
from ._gravity import direct


def all_direct(f: SimSnap, eps: float | SimArray | None = None):
    """Calculate the potential and acceleration for all particles in the snapshot using a direct summation algorithm.

    The results are stored inside the snapshot itself, as f['phi'] and f['acc'].

    .. warning::
       The direct summation algorithm is implemented in Cython and parallelised. Nonetheless, given the O(N^2) scaling
       of the algorithm, it quickly becomes prohibitive for large numbers of particles.

    Parameters
    ----------

    f :
        The snapshot to calculate the potential and acceleration for
    eps :
        The gravitational softening length. If not provided, the value of ``f['eps']`` will be used.

    """
    phi, acc = direct(f, f['pos'].view(np.ndarray), eps)
    f['phi'] = phi
    f['acc'] = acc


def all_pm(f: SimSnap, ngrid: int = 10):
    """Calculate the potential and acceleration for all particles in the snapshot using a Particle-Mesh algorithm.

    The results are stored inside the snapshot itself, as ``f['phi']`` and ``f['acc']``.

    .. warning::
       PM calculations assume periodic boundary conditions, and are only accurate on large scales (much larger than


    Parameters
    ----------

    f :
        The snapshot to calculate the potential and acceleration for

    ngrid :
        The number of grid points to use in each dimension for the Particle-Mesh calculation.

    """
    phi, acc = pm(f, f['pos'].view(np.ndarray), eps, ngrid=ngrid)
    f['phi'] = phi
    f['acc'] = acc


def pm(f: SimSnap, ipos: np.ndarray, ngrid:int = 10, x0=None, x1=None):
    """Calculate the potential and acceleration for a set of particles using a Particle-Mesh algorithm.

    Parameters
    ----------

    f :
        The snapshot to calculate the potential and acceleration for

    ipos :
        The positions of the particles to calculate the potential and acceleration for

    x0 :
        The lower bound of the grid in each dimension. If ``None``, the minimum of the snapshot's positions will be
        used.

    x1 :
        The upper bound of the grid in each dimension. If ``None``, ``x0 + f.properties['boxsize']`` will be used.

    Returns
    -------

    phi : array.SimArray
        The gravitational potential at the specified positions

    grad_phi : array.SimArray
        The gravitational acceleration at the specified positions

    """

    if x0 is None:
        x0 = f['pos'].min()
    if x1 is None:
        x1 = x0 + f.properties['boxsize']

    dx = float(x1 - x0) / ngrid
    grid, edges = np.histogramdd(f['pos'],
                                 bins=ngrid,
                                 range=[(x0, x1), (x0, x1), (x0, x1)],
                                 normed=False,
                                 weights=f['mass'])
    grid /= dx ** 3
    recip_rho_grid = np.fft.rfftn(grid)

    freqs = np.fft.fftfreq(ngrid, d=dx)

    kvecs = np.zeros((ngrid, ngrid, ngrid / 2 + 1, 3))
    kvecs[:, :,:, 0] = freqs.reshape((1, ngrid, 1, 1))
    kvecs[:, :,:, 1] = freqs.reshape((1, 1, ngrid, 1))
    kvecs[:, :,:, 2] = abs(freqs[:ngrid/2+1].reshape((1, 1, 1, ngrid/2+1)))

    k = (kvecs ** 2).sum(axis=3)
    assert k.shape == recip_rho_grid.shape

    recip_phi_grid = 4 * math.pi * recip_rho_grid / k ** 2
    recip_phi_grid[np.where(k == 0)] = 0

    phi_grid = np.fft.irfftn(recip_phi_grid, grid.shape)
    grad_phi_grid = np.concatenate((np.fft.irfftn(-1.j*kvecs[:, :,:, 0]*recip_phi_grid, grid.shape)[:,:,:, np.newaxis],
                                    np.fft.irfftn(-1.j*kvecs[:, :,:, 1]*recip_phi_grid, grid.shape)[:,:,:, np.newaxis],
                                    np.fft.irfftn(-1.j*kvecs[:, :,:, 2]*recip_phi_grid, grid.shape)[:,:,:, np.newaxis]),
                                   axis=3)

    ipos_I = np.array((ipos - x0) / dx, dtype=int)

    phi = np.array([phi_grid[x, y, z] for x, y, z in ipos_I])
    grad_phi = np.array([grad_phi_grid[x, y, z, :] for x, y, z in ipos_I])

    phi = phi.view(array.SimArray)
    phi.units = units.G * f['mass'].units / f['pos'].units

    grad_phi = grad_phi.view(array.SimArray)
    grad_phi.units = units.G * f['mass'].units / f['pos'].units ** 2

    return phi, -grad_phi

def midplane_rot_curve(f: SimSnap, rxy_points: np.ndarray, eps: float | SimArray | None = None):
    """Calculate the rotation curve of a disk galaxy in the x-y midplane (with z=0)

    Parameters
    ----------
    f :
        The snapshot to calculate the rotation curve for
    rxy_points :
        A list or array of radii at which to calculate the rotation curve, in the xy-plane

    Returns
    -------

    v : array.SimArray
        The rotation curve at the specified radii
    """

    if eps is None:
        eps = get_eps(f)
    elif isinstance(eps, (str, units.UnitBase)):
        eps = eps_as_simarray(f, eps)

    # u_out = (units.G * f['mass'].units / f['pos'].units)**(1,2)

    # Do four samples like Tipsy does
    rs = [pos for r in rxy_points for pos in [
        (r, 0, 0), (0, r, 0), (-r, 0, 0), (0, -r, 0)]]

    pot, accel = direct(f, np.array(rs, dtype=f['pos'].dtype), eps=eps)

    u_out = (accel.units * f['pos'].units) ** (1, 2)

    # accel = array.SimArray(m_by_r2,units.G * f['mass'].units / (f['pos'].units**2) )

    vels = []

    i = 0
    for r in rxy_points:
        r_acc_r = []
        for pos in [(r, 0, 0), (0, r, 0), (-r, 0, 0), (0, -r, 0)]:
            r_acc_r.append(np.dot(-accel[i, :], pos))
            i = i + 1

        vel2 = np.mean(r_acc_r)
        if vel2 > 0:
            vel = math.sqrt(vel2)
        else:
            vel = 0

        vels.append(vel)

    x = array.SimArray(vels, units=u_out)
    x.sim = f.ancestor
    return x


def midplane_potential(f, rxy_points, eps=None):
    """Calculate the potential of a disk galaxy in the x-y midplane (with z=0)

    Parameters
    ----------
    f :
        The snapshot to calculate the potential for
    rxy_points :
        A list or array of radii at which to calculate the potential, in the xy-plane

    Returns
    -------

    v : array.SimArray
        The potential at the specified radii
    """

    if eps is None:
        eps = get_eps(f)
    elif isinstance(eps, (str, units.UnitBase)):
        eps = eps_as_simarray(f, eps)

    u_out = units.G * f['mass'].units / f['pos'].units


    # Do four samples like Tipsy does
    rs = [pos for r in rxy_points for pos in [
        (r, 0, 0), (0, r, 0), (-r, 0, 0), (0, -r, 0)]]

    m_by_r, m_by_r2 = direct(f, np.array(rs, dtype=f['pos'].dtype), eps=eps)

    potential = units.G * m_by_r * f['mass'].units / f['pos'].units

    pots = []

    i = 0
    for r in rxy_points:
        # Do four samples like Tipsy does
        pot = []
        for pos in [(r, 0, 0), (0, r, 0), (-r, 0, 0), (0, -r, 0)]:
            pot.append(potential[i])
            i = i + 1

        pots.append(np.mean(pot))

    x = array.SimArray(pots, units=u_out)
    x.sim = f.ancestor
    return x



"""Tools for scientific analysis with pynbody

This sub-package contains a number of modules that enable scientific analysis beyond the basic capabilities of pynbody.
The most essential tools are imported into the sub-package itself, so that they can be accessed directly from
pynbody.analysis. For example, :meth:`pynbody.analysis.halo.center` is accessible as :meth:`pynbody.analysis.center`,
while :meth:`pynbody.analysis.angmom.faceon` is accessible as :meth:`pynbody.analysis.faceon`.

"""

from . import (
    cosmology,
    halo,
    hifrac,
    hmf,
    interpolate,
    ionfrac,
    luminosity,
    pkdgrav_cosmo,
    profile,
    ramses_util,
    theoretical_profiles,
)
from .angmom import faceon, sideon
from .halo import center
from .hmf import halo_mass_function
from .profile import Profile



import numpy as np
import pytest

import pynbody


def test_order_bridge():
    f1 = pynbody.new(dm=1000)
    f2 = pynbody.new(dm=3000)
    f1['iord'] = np.arange(5, 2005, 2, dtype=int)
    f2['iord'] = np.arange(3000, dtype=int)

    b = pynbody.bridge.OrderBridge(f1, f2)

    h1 = f1[:50:2]
    assert b(h1).ancestor is f2
    assert (b(h1)['iord'] == h1['iord']).all


def test_bridge_factory():
    f1 = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")
    f2 = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")
    b = pynbody.bridge.bridge_factory(f1, f2)

    assert isinstance(b, pynbody.bridge.OrderBridge)


def test_nonmonotonic_transfer_matrix():
    # test the case where a non-monotonic bridge between two simulations is queries for particle transfer
    f1 = pynbody.new(dm=10)
    f2 = pynbody.new(dm=10)
    f1['iord'] = np.arange(0,10)
    f2['iord'] = np.array([0,2,4,6,8,1,3,5,7,9],dtype=np.int32)


    # 100% of mass in group 1 transfers to group 1
    # 80% of mass in group 0 transfers to group 0
    # 20% of mass in group 0 transfers to group 1

    f1['grp'] = np.array([0,0,0,0,0,1,1,1,1,1],dtype=np.int32)
    f2['grp'] = np.array([0,0,0,0,1,1,1,1,1,1],dtype=np.int32)[f2['iord']]

    b = pynbody.bridge.OrderBridge(f1,f2,monotonic=False)

    h1 = pynbody.halo.number_array.HaloNumberCatalogue(f1)
    h2 = pynbody.halo.number_array.HaloNumberCatalogue(f2)

    with pytest.warns(DeprecationWarning):
        xfer = b.catalog_transfer_matrix(0,1,h1,h2)

    assert (xfer==[[4,1],[0,5]]).all()


def test_nonmonotonic_incomplete_bridge():
    # Test the non-monotonic bridge where not all the particles are shared
    f1 = pynbody.new(dm=10)
    f2 = pynbody.new(dm=9)
    f1['iord'] = np.arange(0,10)
    f2['iord'] = np.array([0,2,4,6,8,1,5,7,9])
    b = pynbody.bridge.OrderBridge(f1,f2,monotonic=False)

    assert (b(f1[:5])['iord']==np.array([0,1,2,4])).all()


def test_monotonic_incomplete_bridge():
    # Test the monotonic bridge where not all the particles are shared
    f1 = pynbody.new(dm=10)
    f2 = pynbody.new(dm=9)
    f1['iord'] = np.arange(0, 10)
    f2['iord'] = np.array([0, 1, 2, 4, 5 ,6, 7, 8, 9])
    b = pynbody.bridge.OrderBridge(f1, f2, monotonic=False)

    assert (b(f1[:5])['iord'] == np.array([0, 1, 2, 4])).all()


def test_family_morphing():
    f1 = pynbody.new(dm=5, gas=5)
    f2 = pynbody.new(dm=10)

    # set dm and gas iords separately as it's possible the new command initialises them out of order
    f1.dm['iord'] = np.arange(0,5)
    f1.gas['iord'] = np.arange(5,10)
    f2['iord'] = np.array([0,2,4,6,8,1,3,5,7,9])

    b = pynbody.bridge.OrderBridge(f1,f2,monotonic=False,allow_family_change=True)

    assert (b(f2).dm['iord']==np.array([0,2,4,1,3])).all()
    assert (b(f2).gas['iord'] == np.array([6, 8, 5, 7, 9])).all()


def test_bridging_with_more_families():
    # Test that we can create a group array for snapshots that have a complex family structure,
    # and bridge only with one family (DM). This is necessary for e.g. Tangos linking
    f1 = pynbody.load("testdata/ramses_new_format_cosmo_with_ahf_output_00110")
    g1 = pynbody.halo.ahf.AHFCatalogue(f1, halo_numbers='v1')
    g1.load_all()
    # Work only on one family and create a useless bridge which is enough to break the code
    f1 = f1.dm
    b = pynbody.bridge.OrderBridge(f1, f1)

    with pytest.warns(DeprecationWarning):
        # The line below would fail with IndexError: index is out of bound when called for a single family
        # The family slicing is done wrto to the base ancestor, but the IDs were not offset
        # by family.start and would run over the array. This is now fixed
        mat = b.catalog_transfer_matrix(max_index=5, groups_1=g1, groups_2=g1,
                                        use_family=pynbody.family.dm, only_family=pynbody.family.dm)

    # Now it works and produces a diagonla matrix  since we are mapping the same snapshot
    assert(np.count_nonzero(mat - np.diag(np.diagonal(mat))) == 0)
    assert(mat[0, 0] == 12)

def test_fuzzy_match_only_one_family():
    f = pynbody.new(dm=10, gas=10)
    f['iord'] = np.arange(0,20)
    f['grp'] = np.array([0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1],dtype=np.int32)

    f2 = pynbody.new(dm=10, gas=10)
    f2['iord'] = np.arange(0, 20)
    f2['grp'] = f['grp']
    f2.dm['grp'] = np.array([0,0,0,0,0,1,1,1,2,2],dtype=np.int32)

    h = pynbody.halo.number_array.HaloNumberCatalogue(f)
    h2 = pynbody.halo.number_array.HaloNumberCatalogue(f2)
    b = pynbody.bridge.OrderBridge(f,f2)

    with pytest.warns(DeprecationWarning):
        assert b.fuzzy_match_catalog(use_family=pynbody.family.gas,groups_1=h, groups_2=h2)[1]==[(1, 1.0)]
        assert b.fuzzy_match_catalog(use_family=pynbody.family.dm, groups_1=h, groups_2=h2)[1] == [(1, 0.6), (2, 0.4)]
        assert b.fuzzy_match_catalog(groups_1=h, groups_2=h2)[1] == [(1, 0.8), (2, 0.2)]

        # Test that it also works with only_family:
        assert b.fuzzy_match_catalog(only_family=pynbody.family.gas, groups_1=h, groups_2=h2)[1] == [(1, 1.0)]
        assert b.fuzzy_match_catalog(only_family=pynbody.family.dm, groups_1=h, groups_2=h2)[1] == [(1, 0.6), (2, 0.4)]

@pytest.fixture
def snapshot_pair():
    f1 = pynbody.new(dm=100)
    f1['iord'] = np.arange(0, 100)
    f1['grp'] = np.array([0] * 20 + [1] * 10 + [2] * 30 + [4] * 40, dtype=np.int32)

    f2 = pynbody.new(dm=100)
    order = np.random.permutation(100)
    f2['iord'] = order
    f2['grp'] = f1['grp'][order]
    f2['grp'][f2['grp'] == 0] = 1  # imagine that halo 0 merged with halo 1

    # imagine that halo 4 split into lots of little things so we won't get a match
    f2['grp'][f2['grp'] == 4] = np.arange(4, 44)

    return f1, f2

def test_match_halos(snapshot_pair):

    f1, f2 = snapshot_pair

    b = pynbody.bridge.OrderBridge(f1, f2, monotonic=False)
    h1 = f1.halos()
    h2 = f2.halos()

    matched = b.match_halos(h1,h2)

    assert matched.keys() == {0,1,2,4}
    expected_matches = {0:1,1:1,2:2,4:-1}

    for k,v in expected_matches.items():
        assert matched[k] == v


    matched_on_index = b.match_halos(h1,h2,use_halo_indexes=True)

    for k,v in expected_matches.items():
        k_id = h1.number_mapper.number_to_index(k)
        if v != -1:
            v_id = h2.number_mapper.number_to_index(v)
        else:
            v_id = -1
        assert matched_on_index[k_id] == v_id

    assert matched_on_index[0] == 0


    b = pynbody.bridge.OrderBridge(f2, f1, monotonic=False)
    matched_reverse = b.match_halos(h2,h1)
    assert matched_reverse.keys() == set(f2['grp'])
    assert matched_reverse[1] == 0
    assert matched_reverse[2] == 2
    for i in range(4,44):
        assert matched_reverse[i] == 4

def test_fuzzy_match_halos(snapshot_pair):
    f1, f2 = snapshot_pair

    b = pynbody.bridge.OrderBridge(f1, f2, monotonic=False)
    h1 = f1.halos()
    h2 = f2.halos()

    fuzzy_match = b.fuzzy_match_halos(h1, h2)

    assert fuzzy_match.keys() == {0, 1, 2, 4}

    assert fuzzy_match[0] == [(1, 1.0)]
    assert fuzzy_match[1] == [(1, 1.0)]
    assert fuzzy_match[2] == [(2, 1.0)]

    for i in range(4, 44):
        assert (i, 1. / 40) in fuzzy_match[4]

    b = pynbody.bridge.OrderBridge(f2, f1, monotonic=False)
    fuzzy_match_rev = b.fuzzy_match_halos(h2, h1)

    assert fuzzy_match_rev.keys() == set(f2['grp'])

    assert fuzzy_match_rev[1] == [(0, 2./3), (1, 1./3)]



"""

halo
====

Functions for dealing with and manipulating halos in simulations.


"""

import functools
import logging
import math
import operator
import warnings

import numpy as np

from .. import array, config, filt, transformation, units, util
from . import _com, cosmology, profile

logger = logging.getLogger('pynbody.analysis.halo')


def shrink_sphere_center(sim, r=None, shrink_factor=0.7, min_particles=100,
                         num_threads = None, particles_for_velocity = 0,
                         families_for_velocity = ['dm', 'star']):
    """
    Return the center according to the shrinking-sphere method of Power et al (2003)

    Most users will want to use the higher-level :func:`center` function, which actually performs the centering operation.
    This function is a lower-level interface that calculates the center but does not move the particles.

    Parameters
    ----------

    sim : SimSnap
        The simulation snapshot to center

    r : float | str, optional
        Initial search radius. If None, a rough estimate is used.

    shrink_factor : float, optional
        The amount to shrink the search radius by on each iteration

    min_particles : int, optional
        Minimum number of particles within the search radius. When this number is reached, the search is complete.

    num_threads : int, optional
        Number of threads to use for the calculation. If None, the number of threads is taken from the configuration.

    particles_for_velocity : int, optional
        If > min_particles, a velocity centre is calculated when the number of particles falls below this threshold,
        and returned.

    families_for_velocity : list, optional
        The families to use for the velocity centering. Default is ['dm', 'star'], because gas particles may
        be involved in violent outflows making them a risky choice for velocity centering.

    Returns
    -------
    com : SimArray
        The center of mass of the final sphere

    vel : SimArray
        The center of mass velocity of the final sphere. Only returned if particles_for_velocity > min_particles.

    """

    if num_threads is None:
        num_threads = config['number_of_threads']

    if r is None:

        # use rough estimate for a maximum radius
        # results will be insensitive to the exact value chosen
        r = (sim["x"].max() - sim["x"].min()) / 2

    elif isinstance(r, str) or issubclass(r.__class__, units.UnitBase):
        if isinstance(r, str):
            r = units.Unit(r)
        r = r.in_units(sim['pos'].units, **sim.conversion_context())

    mass = np.asarray(sim['mass'], dtype='double')
    pos = np.asarray(sim['pos'], dtype='double')

    R = _com.shrink_sphere_center(pos, mass, min_particles, particles_for_velocity,
                                  shrink_factor, r, num_threads)

    com, final_radius, velocity_radius = R

    logger.info("Radius for velocity measurement = %s", velocity_radius)
    logger.info("Final SSC=%s", com)

    com_to_return = array.SimArray(com, sim['pos'].units)

    if particles_for_velocity > min_particles:
        fam_filter = functools.reduce(operator.or_, (filt.FamilyFilter(f) for f in families_for_velocity))
        final_sphere = sim[filt.Sphere(velocity_radius, com) & fam_filter]
        logger.info("Particles in sphere = %d", len(final_sphere))
        if len(final_sphere) == 0:
            warnings.warn("Final sphere is empty; cannot return a velocity. This probably implies something is "
                          "wrong with the position centre too.", RuntimeWarning)
            return com_to_return, np.array([0., 0., 0.])
        else:
            vel = final_sphere.mean_by_mass('vel')
            logger.info("Final velocity=%s", vel)
            return com_to_return, vel
    else:
        return com_to_return



def virial_radius(sim, cen=None, overden=178, r_max=None, rho_def='matter'):
    """Calculate the virial radius of the halo centered on the given coordinates.

    The default is here defined by the sphere centered on cen which contains a
    mean density of overden * rho_M_0 * (1+z)^3.

    Parameters
    ----------

    sim : SimSnap
        The simulation snapshot for which to calculate the virial radius.

    cen : array_like, optional
        The center of the halo. If None, the halo is assumed to be already centered.

    overden : float, optional
        The overdensity of the halo. Default is 178.

    r_max : float, optional
        The maximum radius to search for the virial radius. If None, the maximum radius of any
        particle in *sim* is used

    rho_def : str, optional
        The density definition to use.
        Default is 'matter', which uses the matter density at the redshift of the simulation. Alternatively,
        'critical' can be used for the critical density at this redshift.

    Returns
    -------

    float
        The virial radius of the halo in the position units of *sim*.

    """

    if r_max is None:
        r_max = (sim["x"].max() - sim["x"].min())
    else:
        if cen is not None:
            sim = sim[filt.Sphere(r_max, cen)]
        else:
            sim = sim[filt.Sphere(r_max)]

    r_min = 0.0

    if rho_def == 'matter':
       ref_density = sim.properties["omegaM0"] * cosmology.rho_crit(sim, z=0) * (1.0 + sim.properties["z"]) ** 3
    elif rho_def == 'critical':
        ref_density = cosmology.rho_crit(sim, z=sim.properties["z"])
    else:
        raise ValueError(rho_def + "is not a valid definition for the reference density")

    target_rho = overden * ref_density
    logger.info("target_rho=%s", target_rho)

    if cen is not None:
        transform = sim.translate(-np.asanyarray(cen))
    else:
        transform = transformation.NullTransformation(sim)

    with transform:
        sim = sim[filt.Sphere(r_max)]
        with sim.immediate_mode:
            mass_ar = np.asarray(sim['mass'])
            r_ar = np.asarray(sim['r'])

        rho = lambda r: util.sum_if_lt(mass_ar,r_ar,r)/(4. * math.pi * (r ** 3) / 3)
        result = util.bisect(r_min, r_max, lambda r: target_rho -
                             rho(r), epsilon=0, eta=1.e-3 * target_rho, verbose=False)

    return result


def _potential_minimum(sim):
    i = sim["phi"].argmin()
    return sim["pos"][i].copy()


def hybrid_center(sim, r='3 kpc', **kwargs):
    """Determine the center of the halo by finding the shrink-sphere-center near the potential minimum

    Most users will want to use the general :func:`center` function, which actually performs the centering operation.
    This function is a lower-level interface that calculates the center but does not move the particles.

    Parameters
    ----------

    sim : SimSnap
        The simulation snapshot of which to find the center
    r : float | str, optional
        Radius from the potential minimum to search for the center. Default is 3 kpc.

    Remaining parameters are passed onto :func:`shrink_sphere_center`.

    Returns
    -------

    com : SimArray
        The center of mass of the final sphere
    vel : SimArray
        The center of mass velocity of the final sphere. Only returned if particles_for_velocity > min_particles.


    """

    try:
        cen_a = _potential_minimum(sim)
    except KeyError:
        cen_a = center_of_mass(sim)
    return shrink_sphere_center(sim[filt.Sphere(r, cen_a)], **kwargs)



def vel_center(sim, cen_size="1 kpc", return_cen=False, move_all=True, **kwargs):
    """Recenter the snapshot on the center of mass velocity inside a sphere of specified radius

    Attempts to use the star particles, falling back to gas or dark matter if necessary.

    Parameters
    ----------

    sim : SimSnap
        The simulation snapshot to center

    cen_size : str or float, optional
        The size of the sphere to use for the velocity centering. Default is 1 kpc.

    return_cen : bool, optional
        If True, only return the velocity center without actually moving the snapshot. Default is False.

    move_all : bool, optional
        If True (default), move the entire snapshot. Otherwise only move the particles in the halo passed in.

    retcen : bool, optional
        Deprecated alias for return_cen

    Returns
    -------

    Transformation | SimArray
        Normally, a transformation object that can be used to revert the transformation.
        However, if return_cen is True, a SimArray containing the velocity center
        coordinates is returned instead, and the snapshot is not transformed.

    """

    if "retcen" in kwargs:
        return_cen = kwargs.pop("retcen")
        warnings.warn("The 'retcen' keyword is deprecated. Use 'return_cen' instead.", DeprecationWarning)

    logger.info("Finding halo velocity center...")

    if move_all:
        target = sim.ancestor
    else:
        target = sim

    cen = sim.star[filt.Sphere(cen_size)]
    if len(cen) < 5:
        # fall-back to DM
        cen = sim.dm[filt.Sphere(cen_size)]
    if len(cen) < 5:
        # fall-back to gas
        cen = sim.gas[filt.Sphere(cen_size)]
    if len(cen) < 5:
        # very weird snapshot, or mis-centering!
        raise ValueError("Insufficient particles around center to get velocity")

    vcen = (cen['vel'].transpose() * cen['mass']).sum(axis=1) / \
        cen['mass'].sum()
    vcen.units = cen['vel'].units

    logger.info("vcen=%s", vcen)

    if return_cen:
        return vcen
    else:
        return target.offset_velocity(-vcen)


def center(sim, mode=None, return_cen=False, with_velocity=True, cen_size="1 kpc",
           cen_num_particles=10000, move_all=True, wrap=False, **kwargs):
    """Transform the ancestor snapshot so that the provided snapshot is centred

    The centering scheme is determined by the ``mode`` keyword. As well as the
    position, the velocity can also be centred.

    The following centring modes are available:

    *  *pot*: potential minimum

    *  *com*: center of mass

    *  *ssc*: shrink sphere center

    *  *hyb*: for most halos, returns the same as ssc, but works faster by starting iteration near potential minimum

    Before the main centring routine is called, the snapshot is translated so that the
    halo is already near the origin. The box is then wrapped so that halos on the edge
    of the box are handled correctly.


    Parameters
    ----------

    sim : SimSnap
        The simulation snapshot from which to derive a centre. The ancestor snapshot is
        then transformed.

    mode : str or function, optional
        The method to use to determine the centre. If None, the default is taken from the configuration.
        Accepted values are discussed above. A function returning the centre, or a pair of
        centres (position and velocity) can also be passed.

    cen_size : str or float, optional
        The size of the sphere to use for the velocity centering. Default is 1 kpc.
        Note that this is only used if velocity centring is requested but the underlying
        method does not return a velocity centre. For example, if using the 'ssc' method,
        the cen_num_particles keyword should be used instead.

    cen_num_particles : int, optional
        The number of particles to use for the velocity centering. Default is 5000.
        This is passed to the 'ssc' method, which then finds the sphere with approximately
        this number of particles in it for the velocity centering.

    with_velocity: bool, optional
        If True, also center the velocity. Default is True.

    return_cen: bool, optional
        If True, only return the center without actually centering the snapshot.
        Default is False.

    move_all: bool, optional
        If True (default), move the entire snapshot. Otherwise only move the particles
        in the halo/subsnap passed into this function.

    vel: bool, optional
        Deprecated alias for with_velocity. Default is True.

    retcen: bool, optional
        If True, only return the center without centering the snapshot. Default is False.

    Returns
    -------
    Transformation | SimArray
        Normally, a transformation object that can be used to revert the transformation.
        However, if return_cen is True, a SimArray containing the center
        coordinates is returned instead, and the snapshot is not transformed.



    """

    if 'vel' in kwargs:
        warnings.warn("The 'vel' keyword is deprecated. Use 'with_velocity' instead.", DeprecationWarning)
        with_velocity = kwargs.pop('vel')

    if 'retcen' in kwargs:
        warnings.warn("The 'retcen' keyword is deprecated. Use 'return_cen' instead.", DeprecationWarning)
        return_cen = kwargs.pop('retcen')


    global config
    if mode is None:
        mode = config['centering-scheme']

    try:
        fn = {'pot': _potential_minimum,
              'com': lambda s : s.mean_by_mass('pos'),
              'ssc': functools.partial(shrink_sphere_center, particles_for_velocity=cen_num_particles),
              'hyb': hybrid_center}[mode]
    except KeyError:
        fn = mode

    if move_all:
        target = sim.ancestor
    else:
        target = sim

    if wrap:
        # centre on something within the halo and wrap
        initial_offset = -sim['pos'][0]
        transform = target.translate(initial_offset)
        target.wrap()
    else:
        transform = transformation.NullTransformation(target)
        initial_offset = np.array([0., 0., 0.])

    try:
        centre = fn(sim, **kwargs)
        if len(centre) == 2:
            # implies we have a velocity centre as well
            centre, vel_centre = centre
        else:
            vel_centre = None

        if return_cen:
            transform.revert()
            return centre - initial_offset

        transform = transform.translate(-centre)

        if with_velocity:
            if vel_centre is None :
                vel_centre = vel_center(sim, cen_size=cen_size, retcen=True)
            logger.info("vel_centre=%s", vel_centre)
            transform = transform.offset_velocity(-vel_centre)

    except:
        transform.revert()
        raise

    return transform

def halo_shape(sim, N=100, rin=None, rout=None, bins='equal'):
    """
    Computes the shape of a halo as a function of radius by fitting homeoidal shells.

    The halo must be pre-centred, e.g. using :func:`center`.

    Caution is advised when assigning large number of bins and radial
    ranges with many particles, as the algorithm becomes very slow.

    Parameters
    ----------

    N : int
        The number of homeoidal shells to consider. Shells with few particles will take longer to fit.

    rin : float
        The minimum radial bin in units of sim['pos']. By default this is taken as rout/1000.
        Note that this applies to axis a, so particles within this radius may still be included within
        homeoidal shells.

    rout : float
        The maximum radial bin in units of sim['pos']. By default this is taken as the largest radial value
        in the halo particle distribution.

    bins : str
        The spacing scheme for the homeoidal shell bins. 'equal' initialises radial bins with equal numbers
        of particles, with the exception of the final bin which will accomodate remainders. This
        number is not necessarily maintained during fitting. 'log' and 'lin' initialise bins
        with logarithmic and linear radial spacing.

    Returns
    -------

    rbin : SimArray
        The radial bins used for the fitting.

    ba : array
        The axial ratio b/a as a function of radius.

    ca : array
        The axial ratio c/a as a function of radius.

    angle : array
        The angle of the a-direction with respect to the x-axis as a function of radius.

    Es : array
        The rotation matrices for each shell.

    """

    #-----------------------------FUNCTIONS-----------------------------
    # Define an ellipsoid shell with lengths a,b,c and orientation E:
    def Ellipsoid(r, a,b,c, E):
        x,y,z = np.dot(np.transpose(E),[r[:,0],r[:,1],r[:,2]])
        return (x/a)**2 + (y/b)**2 + (z/c)**2

    # Define moment of inertia tensor:
    MoI = lambda r,m: np.array([[np.sum(m*r[:,i]*r[:,j]) for j in range(3)]\
                               for i in range(3)])

    # Splits 'r' array into N groups containing equal numbers of particles.
    # An array is returned with the radial bins that contain these groups.
    sn = lambda r,N: np.append([r[i*int(len(r)/N):(1+i)*int(len(r)/N)][0]\
                               for i in range(N)],r[-1])

    # Retrieves alignment angle:
    almnt = lambda E: np.arccos(np.dot(np.dot(E,[1.,0.,0.]),[1.,0.,0.]))
    #-----------------------------FUNCTIONS-----------------------------

    if (rout is None): rout = sim.dm['r'].max()
    if (rin is None): rin = rout/1E3

    posr = np.array(sim.dm['r'])[np.where(sim.dm['r'] < rout)]
    pos = np.array(sim.dm['pos'])[np.where(sim.dm['r'] < rout)]
    mass = np.array(sim.dm['mass'])[np.where(sim.dm['r'] < rout)]

    rx = [[1.,0.,0.],[0.,0.,-1.],[0.,1.,0.]]
    ry = [[0.,0.,1.],[0.,1.,0.],[-1.,0.,0.]]
    rz = [[0.,-1.,0.],[1.,0.,0.],[0.,0.,1.]]

    # Define bins:
    if (bins == 'equal'): # Each bin contains equal number of particles
        mid = sn(np.sort(posr[np.where((posr >= rin) & (posr <= rout))]),N*2)
        rbin = mid[1:N*2+1:2]
        mid = mid[0:N*2+1:2]

    elif (bins == 'log'): # Bins are logarithmically spaced
        mid = profile.Profile(sim.dm, type='log', ndim=3, rmin=rin, rmax=rout, nbins=N+1)['rbins']
        rbin = np.sqrt(mid[0:N]*mid[1:N+1])

    elif (bins == 'lin'): # Bins are linearly spaced
        mid = profile.Profile(sim.dm, type='lin', ndim=3, rmin=rin, rmax=rout, nbins=N+1)['rbins']
        rbin = 0.5*(mid[0:N]+mid[1:N+1])

    # Define b/a and c/a ratios and angle arrays:
    ba,ca,angle = np.zeros(N),np.zeros(N),np.zeros(N)
    Es = [0]*N

    # Begin loop through radii:
    for i in range(0,N):

        # Initialise convergence criterion:
        tol = 1E-3
        count = 0

        # Define initial spherical shell:
        a=b=c = rbin[i]
        E = np.identity(3)
        L1,L2 = rbin[i]-mid[i],mid[i+1]-rbin[i]

        # Begin iterative procedure to fit data to shell:
        while True:
            count += 1

            # Collect all particle positions and masses within shell:
            r = pos[np.where((posr < a+L2) & (posr > c-L1*c/a))]
            inner = Ellipsoid(r, a-L1,b-L1*b/a,c-L1*c/a, E)
            outer = Ellipsoid(r, a+L2,b+L2*b/a,c+L2*c/a, E)
            r = r[np.where((inner > 1.) & (outer < 1.))]
            m = mass[np.where((inner > 1.) & (outer < 1.))]

            # End iterations if there is no data in range:
            if (len(r) == 0):
                ba[i],ca[i],angle[i],Es[i] = b/a,c/a,almnt(E),E
                logger.info('No data in range after %i iterations' %count)
                break

            # Calculate shape tensor & diagonalise:
            D = list(np.linalg.eig(MoI(r,m)/np.sum(m)))

            # Purge complex numbers:
            if isinstance(D[1][0,0],complex):
                D[0] = D[0].real ; D[1] = D[1].real
                logger.info('Complex numbers in D removed...')

            # Compute ratios a,b,c from moment of intertia principles:
            anew,bnew,cnew = np.sqrt(abs(D[0])*3.0)

            # The rotation matrix must be reoriented:
            E = D[1]
            if ((bnew > anew) & (anew >= cnew)): E = np.dot(E,rz)
            if ((cnew > anew) & (anew >= bnew)): E = np.dot(np.dot(E,ry),rx)
            if ((bnew > cnew) & (cnew >= anew)): E = np.dot(np.dot(E,rz),rx)
            if ((anew > cnew) & (cnew >= bnew)): E = np.dot(E,rx)
            if ((cnew > bnew) & (bnew >= anew)): E = np.dot(E,ry)
            cnew,bnew,anew = np.sort(np.sqrt(abs(D[0])*3.0))

            # Keep a as semi-major axis and distort b,c by b/a and c/a:
            div = rbin[i]/anew
            anew *= div
            bnew *= div
            cnew *= div

            # Convergence criterion:
            if (np.abs(b/a-bnew/anew) < tol) & (np.abs(c/a-cnew/anew) < tol):
                if (almnt(-E) < almnt(E)): E = -E
                ba[i],ca[i],angle[i],Es[i] = bnew/anew,cnew/anew,almnt(E),E
                break

            # Increase tolerance if convergence has stagnated:
            elif (count%10 == 0): tol *= 5.

            # Reset a,b,c for the next iteration:
            a,b,c = anew,bnew,cnew

    return [array.SimArray(rbin, sim.d['pos'].units), ba, ca, angle, Es]



"""Support for the Rockstar halo finder"""

from __future__ import annotations

import glob
import os.path
import sys

import numpy as np

from .. import util
from . import HaloCatalogue
from .details import number_mapping


class RockstarFormatRevisionError(RuntimeError):
    pass


class RockstarCatalogue(HaloCatalogue):
    def __init__(self, sim, filename=None, format_revision=None):
        """Initialize a RockstarCatalogue.

        Parameters
        ----------

        sim : :class:`pynbody.snapshot.simsnap.SimSnap`
            The simulation snapshot to which this catalogue applies.

        filename : str, optional
            The path to the Rockstar outputs. This can either be the first binary file in the series (e.g.
            ``path/to/halos_15.0.bin``) or the stem for the series (e.g. ``path/to/halos_15``). If not specified,
            the code will attempt to find the Rockstar outputs in the same directory as the simulation snapshot.

        format_revision : int, str, optional
            Override the header's format revision information. Specify 1, 2, 'caterpillar', 'galaxies' for Rockstar
            prior to 2014, post 2014, customized for the caterpillar project and for rockstar galaxies respectively.
            If not specified, the code will attempt to detect the format revision automatically.

        """

        if filename is not None:
            filestem = str(filename)
            if filestem.endswith(".0.bin"):
                filestem = filestem[:-6]
            self._files = glob.glob(filestem + "*.bin") + glob.glob(filestem + "*.boundbin")
        else:
            pathname = os.path.dirname(sim.filename)
            self._files = glob.glob(os.path.join(pathname, 'halos*.bin'))
            if len(self._files)==0 :
                self._files = glob.glob(os.path.join(pathname, 'halos*.boundbin'))
            self._files.sort()

        if len(self._files)==0:
            raise OSError("Could not find any Rockstar output. Try specifying pathname='/path/to/rockstar/outputfolder'")

        self._cpus = [_RockstarCatalogueOneCpu(file_i, format_revision=format_revision) for file_i in self._files]
        self._prune_files_from_wrong_scalefactor(sim)

        halomins = [cpu.halo_min_inclusive for cpu in self._cpus]
        # order self._cpus and self._files by halomins
        # this isn't technically necessary but it means that the internal halo indexing will be
        # in the same order as the halo numbering in the files. In fact, it probably means that
        # the internal halo indexing will be identical to the halo numbering in the files, but we
        # won't take that for granted anywhere.
        self._cpus = [x for _, x in sorted(zip(halomins, self._cpus))]
        self._files = [x for _, x in sorted(zip(halomins, self._files))]

        halo_numbers = np.empty(sum(len(x) for x in self._cpus), dtype=int)
        self._cpu_per_halo = np.empty(len(halo_numbers), dtype=int)
        i = 0
        for j, cat in enumerate(self._cpus):
            halo_numbers[i:i+len(cat)] = cat.get_halo_numbers()
            self._cpu_per_halo[i:i+len(cat)] = j
            i+=len(cat)
        assert i == len(halo_numbers)

        super().__init__(sim, number_mapping.create_halo_number_mapper(halo_numbers))



    def _prune_files_from_wrong_scalefactor(self, sim):
        new_cpus = []
        new_files = []
        for file,cpu in zip(self._files,self._cpus):
            if abs(sim.properties['a']-cpu._head['scale'][0])<1e-6:
                new_cpus.append(cpu)
                new_files.append(file)
        self._cpus = new_cpus
        self._files = new_files

    def _get_particle_indices_one_halo(self, halo_number):
        halo_index = self.number_mapper.number_to_index(halo_number)
        cpu = self._cpu_per_halo[halo_index]
        iords = self._cpus[cpu].read_iords_for_halo(halo_number)
        self._init_iord_to_fpos()
        return self._iord_to_fpos.map_ignoring_order(iords)

    def _get_all_particle_indices(self):
        iords = np.empty(0, dtype=int)
        boundaries = np.empty((0, 2), dtype=int)
        for cpu in self._cpus:
            iords_this_cpu, boundaries_this_cpu = cpu.read_iords_for_all_halos()
            boundaries = np.append(boundaries, boundaries_this_cpu + len(iords), axis=0)
            iords = np.append(iords, iords_this_cpu)

        self._init_iord_to_fpos()
        fpos = iords # nb this doesn't copy! but we won't use the iords again after this, so it's ok
        for a, b in boundaries:
            # iord_to_fpos may not retain ordering, so have to do this per halo
            fpos[a:b] = self._iord_to_fpos.map_ignoring_order(iords[a:b])
        return fpos, boundaries

    def get_properties_one_halo(self, halo_number):
        halo_index = self.number_mapper.number_to_index(halo_number)
        cpu = self._cpu_per_halo[halo_index]
        return self._cpus[cpu].read_properties_for_halo(halo_number)

    def get_properties_all_halos(self, with_units=True) -> dict:
        props = {}
        for cpu in self._cpus:
            props_this_cpu = cpu.read_properties_all_halos()
            # concatenate all the properties
            for k, v in props_this_cpu.items():
                if k in props:
                    props[k] = np.concatenate((props[k], v))
                else:
                    props[k] = v

        return props

    @classmethod
    def _can_load(cls, sim, filename=None, format_revision=None):
        if filename is None:
            return len(
                glob.glob(os.path.join(os.path.dirname(sim.filename), 'halos*.bin'))
            ) > 0
        else:
            if str(filename).endswith(".0.bin") and os.path.exists(filename):
                return True
            if os.path.exists(str(filename)+".0.bin"):
                return True
            return False



class _RockstarCatalogueOneCpu:
    """
    Low-level reader for single CPU output from Rockstar. Users should normally not use this class,
    rather using RockstarCatalogue which collates the multiple sub-files that Rockstar produces.
    """

    head_type = np.dtype([('magic',np.uint64),('snap',np.int64),
                          ('chunk',np.int64),('scale','f'),
                          ('Om','f'),('Ol','f'),('h0','f'),
                          ('bounds','f',6),('num_halos',np.int64),
                          ('num_particles',np.int64),('box_size','f'),
                          ('particle_mass','f'),('particle_type',np.int64),
                          ('format_revision',np.int32),
                          ('rockstar_version',np.str_,12)])

    halo_types = {
        1: np.dtype([('id', np.int64), ('pos', 'f', 3), ('vel', 'f', 3),
                  ('corevel', 'f', 3), ('bulkvel', 'f', 3), ('m', 'f'),
                  ('r', 'f'),
                  ('child_r', 'f'), ('vmax_r', 'f'), ('mgrav', 'f'),
                  ('vmax', 'f'), ('rvmax', 'f'), ('rs', 'f'),
                  ('klypin_rs', 'f'), ('vrms', 'f'), ('J', 'f', 3),
                  ('energy', 'f'), ('spin', 'f'), ('alt_m', 'f', 4),
                  ('Xoff', 'f'), ('Voff', 'f'), ('b_to_a', 'f'),
                  ('c_to_a', 'f'), ('A', 'f', 3), ('b_to_a2', 'f'),
                  ('c_to_a2', 'f'), ('A2', 'f', 3), ('bullock_spin', 'f'),
                  ('kin_to_pot', 'f'), ('m_pe_b', 'f'), ('m_pe_d', 'f'),
                  ('num_p', np.int64), ('num_child_particles', np.int64),
                  ('p_start', np.int64), ('desc', np.int64),
                  ('flags', np.int64), ('n_core', np.int64),
                  ('min_pos_err', 'f'), ('min_vel_err', 'f'),
                  ('min_bulkvel_err', 'f')], align=True)  # Rockstar format v1
        ,
        2: np.dtype([('id',np.int64),('pos','f',3),('vel','f',3),
                          ('corevel','f',3),('bulkvel','f',3),('m','f'),
                          ('r','f'),
                          ('child_r','f'),('vmax_r','f'),('mgrav','f'),
                          ('vmax','f'),('rvmax','f'),('rs','f'),
                          ('klypin_rs','f'),('vrms','f'),('J','f',3),
                          ('energy','f'),('spin','f'),('alt_m','f',4),
                          ('Xoff','f'),('Voff','f'),('b_to_a','f'),
                          ('c_to_a','f'),('A','f',3),('b_to_a2','f'),
                          ('c_to_a2','f'),('A2','f',3),('bullock_spin','f'),
                          ('kin_to_pot','f'),('m_pe_b','f'),('m_pe_d','f'),
                          ('halfmass_radius','f'),
                          ('num_p',np.int64),('num_child_particles',np.int64),
                          ('p_start',np.int64),('desc',np.int64),
                          ('flags',np.int64),('n_core',np.int64),
                          ('min_pos_err','f'),('min_vel_err','f'),
                          ('min_bulkvel_err','f')], align=True), # Rockstar format v2, includes halfmass_radius

        'caterpillar': np.dtype([('id',np.int64),
                                 ('pos','f',3),('vel','f',3),
                          ('corevel','f',3),('bulkvel','f',3),('m','f'),
                          ('r','f'),
                          ('child_r','f'),('vmax_r','f'),('mgrav','f'),
                          ('vmax','f'),('rvmax','f'),('rs','f'),
                          ('klypin_rs','f'),('vrms','f'),('J','f',3),
                          ('energy','f'),('spin','f'),('alt_m','f',4),
                          ('Xoff','f'),('Voff','f'),('b_to_a','f'),
                          ('c_to_a','f'),('A','f',3),('b_to_a2','f'),
                          ('c_to_a2','f'),('A2','f',3),('bullock_spin','f'),
                          ('kin_to_pot','f'),('m_pe_b','f'),('m_pe_d','f'),
                          ('halfmass_radius','f'),
                          ('num_p',np.int64),('num_child_particles',np.int64),

                          ('p_start',np.int64),('desc',np.int64),
                          ('flags',np.int64),('n_core',np.int64),
                          ('min_pos_err','f'),('min_vel_err','f'),
                          ('min_bulkvel_err','f'),
                          ('num_bound', 'i8'), ('num_iter', 'i8')]
                                , align=True), # Hacked rockstar from caterpillar project
        'galaxies':  np.dtype(
        [
                ("id", np.int64),
                ("pos", np.float32, 3),
                ("vel", np.float32, 3),
                ("corevel", np.float32, 3),
                ("bulkvel", np.float32, 3),
                ("m", np.float32),
                ("r", np.float32),
                ("child_r", np.float32),
                ("vmax_r", np.float32),
                ("mgrav", np.float32),
                ("vmax", np.float32),
                ("rvmax", np.float32),
                ("rs", np.float32),
                ("klypin_rs", np.float32),
                ("vrms", np.float32),
                ("J", np.float32, 3),
                ("energy", np.float32),
                ("spin", np.float32),
                ("alt_m", np.float32, 4),
                ("Xoff", np.float32),
                ("Voff", np.float32),
                ("b_to_a", np.float32),
                ("c_to_a", np.float32),
                ("A", np.float32, 3),
                ("b_to_a2", np.float32),
                ("c_to_a2", np.float32),
                ("A2", np.float32, 3),
                ("bullock_spin", np.float32),
                ("kin_to_pot", np.float32),
                ("m_pe_b", np.float32),
                ("m_pe_d", np.float32),
                ("num_p", np.int64),
                ("num_child_particles", np.int64),
                ("p_start", np.int64),
                ("desc", np.int64),
                ("flags", np.int64),
                ("n_core", np.int64),
                ("min_pos_err", np.float32),
                ("min_vel_err", np.float32),
                ("min_bulkvel_err", np.float32),
                ("type", np.int32),
                ("sm", np.float32),
                ("gas", np.float32),
                ("bh", np.float32),
                ("peak_density", np.float32),
                ("av_density", np.float32),
            ],
            align=True
        ), # Galaxy format from Rockstar
    }


    def __init__(self, filename=None, format_revision=None):

        self._rsFilename = filename


        if not os.path.exists(self._rsFilename):
            raise OSError(
                "Halo catalogue not found -- check the file name of catalogue data"
                " or try specifying a catalogue using the filename keyword"
            )

        with util.open_(self._rsFilename, 'rb') as f:
            self._head = np.frombuffer(f.read(self.head_type.itemsize),
                                       dtype=self.head_type)

            # Seek to absolute position
            f.seek(256)

            self._nhalos = self._head['num_halos'][0]

            self._load_rs_halos_with_format_detection(f, format_revision)

    def _load_rs_halos_with_format_detection(self, f, format_revision):
        if format_revision is None:
            # The 'galaxies' format can be either 1 or 2, so we need to try it
            # in both cases.
            format_revision_to_try = [self._head['format_revision'][0], "galaxies"]
        else:
            format_revision_to_try = [format_revision]

        current_pos = f.tell()
        for format_revision in format_revision_to_try:
            f.seek(current_pos)
            try:
                self.halo_type = self.halo_types[format_revision]
                self._load_rs_halos(f)
                return
            except RockstarFormatRevisionError:
                pass

        raise RockstarFormatRevisionError(
            "Could not detect the format revision of the Rockstar catalogue."
        )


    def __len__(self):
        return len(self._halo_lens)



    def read_properties_for_halo(self, n):
        if n<self.halo_min_inclusive or n>=self.halo_max_exclusive:
            raise KeyError("No such halo")

        with util.open_(self._rsFilename, 'rb') as f:
            f.seek(self._haloprops_offset + (n - self.halo_min_inclusive) * self.halo_type.itemsize)
            halo_data = np.fromfile(f, dtype=self.halo_type, count=1)

        # TODO: properties are in Msun / h, Mpc / h
        return dict(list(zip(halo_data.dtype.names,halo_data[0])))

    def read_properties_all_halos(self):
        with util.open_(self._rsFilename, 'rb') as f:
            f.seek(self._haloprops_offset)
            data = np.fromfile(f, dtype=self.halo_type, count=self.halo_max_exclusive - self.halo_min_inclusive)

        data_dict = {name: data[name] for name in data.dtype.names}

        return data_dict

    def _load_rs_halos(self, f):
        self._haloprops_offset = f.tell()
        self._halo_offsets = np.empty(self._head['num_halos'][0],dtype=np.int64)
        self._halo_lens = np.empty(self._head['num_halos'][0],dtype=np.int64)

        offset = self._haloprops_offset+self.halo_type.itemsize*self._head['num_halos'][0]

        self.halo_min_inclusive = int(np.fromfile(f, dtype=self.halo_type, count=1)['id'][0])
        self.halo_max_exclusive = int(self.halo_min_inclusive + self._head['num_halos'][0])

        f.seek(self._haloprops_offset)

        for this_id in range(self.halo_min_inclusive, self.halo_max_exclusive):
            halo_data = np.fromfile(f, dtype=self.halo_type, count=1)
            if halo_data['id'] != this_id:
                raise RockstarFormatRevisionError(
                    "Error while reading halo catalogue. Expected "
                    "halo ID %d, but got %d" % (this_id, halo_data['id'])
                )
            self._halo_offsets[this_id - self.halo_min_inclusive] = offset
            if 'num_bound' in self.halo_type.names:
                num_ptcls = int(halo_data['num_bound'][0])
            else:
                num_ptcls = int(halo_data['num_p'][0])
            self._halo_lens[this_id - self.halo_min_inclusive] = num_ptcls
            offset+=num_ptcls*np.dtype('int64').itemsize


    def get_halo_numbers(self):
        return np.arange(self.halo_min_inclusive, self.halo_max_exclusive)

    def read_iords_for_halo(self, num):
        if num<self.halo_min_inclusive or num>=self.halo_max_exclusive:
            raise KeyError("No such halo")

        with util.open_(self._rsFilename, 'rb') as f:
            f.seek(self._halo_offsets[num - self.halo_min_inclusive])
            return np.fromfile(f, dtype=np.int64, count=self._halo_lens[num - self.halo_min_inclusive])

    def read_iords_for_all_halos(self):
        """Returns an array with all halo iords, and the boundaries of each halo in the array."""
        # check the halo ids are contiguous as expected
        assert (np.diff(self._halo_offsets) == 8 * self._halo_lens[:-1]).all()
        with util.open_(self._rsFilename, 'rb') as f:
            f.seek(self._halo_offsets[0])
            iords = np.fromfile(f, dtype=np.int64, count=self._halo_lens.sum())

        boundaries = np.empty((len(self._halo_lens), 2), dtype=np.int64)
        boundaries[:,1] = np.cumsum(self._halo_lens)
        boundaries[0,0] = 0
        boundaries[1:,0] = boundaries[:-1,1]

        return iords, boundaries



from abc import ABC, abstractmethod
from typing import Union

import numpy as np
from numpy import typing as npt


class HaloNumberMapper(ABC):

    @abstractmethod
    def number_to_index(self, halo_number: Union[int, npt.NDArray[int]]) -> Union[int, npt.NDArray[int]]:
        """Convert a halo number to a zero-based index """
        pass

    @abstractmethod
    def index_to_number(self, halo_index: Union[int, npt.NDArray[int]]) -> Union[int, npt.NDArray[int]]:
        """Convert a zero-based offset to a halo number"""
        pass

    @abstractmethod
    def __len__(self):
        """Returns the number of halos in the catalogue"""
        pass

    @abstractmethod
    def __iter__(self):
        """Iterates over the available halo numbers"""
        pass

    def __repr__(self):
        return f"<{self.__class__.__name__} len={len(self)}>"

    @property
    @abstractmethod
    def all_numbers(self):
        pass


class MonotonicHaloNumberMapper(HaloNumberMapper):
    def __init__(self, halo_numbers: npt.NDArray[int]):
        """A HaloNumberMapper maps halo numbers (arbitrary unique integers) to zero-based indices within a halo catalogue.

        The halo_numbers array must be monotonically increasing.
        """
        halo_numbers = np.asarray(halo_numbers)
        assert np.all(np.diff(halo_numbers) > 0)
        self._halo_numbers = halo_numbers

    def number_to_index(self, halo_number: Union[int, npt.NDArray[int]]) -> Union[int, npt.NDArray[int]]:
        """Convert a halo number to a zero-based index """
        halo_number = np.asarray(halo_number)
        halo_index = np.searchsorted(self._halo_numbers, halo_number)
        if isinstance(halo_index, np.ndarray):
            missing_halo_mask = (halo_index >= len(self._halo_numbers)) | (self._halo_numbers[halo_index] != halo_number)
            missing_halo_numbers = halo_number[missing_halo_mask]

            if missing_halo_numbers.size > 0:
                raise KeyError(f"No such halos: {missing_halo_numbers}")
        else:
            if halo_index >= len(self._halo_numbers) or self._halo_numbers[halo_index] != halo_number:
                raise KeyError(f"No such halo {halo_number}")
        return halo_index

    def index_to_number(self, halo_index: Union[int, npt.NDArray[int]]) -> Union[int, npt.NDArray[int]]:
        """Convert a zero-based offset to a halo number"""
        return self._halo_numbers[halo_index]

    def __len__(self):
        """Returns the number of halos in the catalogue"""
        return len(self._halo_numbers)

    def __iter__(self):
        """Iterates over the available halo numbers"""
        yield from self._halo_numbers

    @property
    def all_numbers(self):
        return self._halo_numbers


class SimpleHaloNumberMapper(HaloNumberMapper):
    def __init__(self, zero_offset, num_halos):
        """A HaloNumberMapper where the relationship between halo numbers and indices is simply an offset"""
        self.zero_offset = zero_offset
        self.num_halos = num_halos

    def number_to_index(self, halo_number: Union[int, npt.NDArray[int]]) -> Union[int, npt.NDArray[int]]:
        if hasattr(halo_number, "__len__"):
            halo_number = np.asarray(halo_number)
            index = halo_number - self.zero_offset
            missing_halo_mask = (index < 0) | (index >= self.num_halos)
            missing_halo_numbers = np.asarray(halo_number)[missing_halo_mask]

            if missing_halo_numbers.size > 0:
                raise KeyError(f"No such halos: {missing_halo_numbers}")
        else:
            index = halo_number - self.zero_offset
            if index< 0 or index >= self.num_halos:
                raise KeyError(f"No such halo {halo_number}")
        return index

    def index_to_number(self, halo_index: Union[int, npt.NDArray[int]]) -> Union[int, npt.NDArray[int]]:
        if hasattr(halo_index, "__len__"):
            halo_index = np.asarray(halo_index)
            missing_halo_mask = (halo_index < 0) | (halo_index >= self.num_halos)
            missing_halo_index = halo_index[missing_halo_mask]

            if missing_halo_index.size > 0:
                raise IndexError(f"No such halos: {missing_halo_index}")
        else:
            if halo_index < 0 or halo_index >= self.num_halos:
                raise IndexError(f"No such halo {halo_index}")

        return halo_index + self.zero_offset

    def __len__(self):
        return self.num_halos

    def __iter__(self):
        yield from range(self.zero_offset, self.zero_offset + self.num_halos)

    @property
    def all_numbers(self):
        return np.arange(self.zero_offset, self.zero_offset + self.num_halos)

class NonMonotonicHaloNumberMapper(MonotonicHaloNumberMapper):
    def __init__(self, halo_numbers_or_ordering: npt.NDArray[int], ordering=False, start_index=0):
        """A HaloNumberMapper that allows for arbitrary mappings between halo numbers and indices.

        Can be created either by specifying the halo numbers for each halo, in index order, or by specifying
        the ordering that the halos should take, starting from start_index and running contiguously.

        Parameters
        ----------

        halo_numbers_or_ordering : array-like
            If ordering is False, this should be an array of halo numbers. If ordering is True, this should be an array
            of the same length as the number of halos, specifying the order in which the halos should be presented.
        ordering: bool
            The meaning of halo_numbers_or_ordering. If False, it is an array of halo numbers. If True, it is an array
            of the same length as the number of halos, specifying the order in which the halos should be presented,
            starting at start_index.
        start_index: int
            The starting index for the ordering. Only used if ordering is True.
        """

        if not ordering and start_index != 0:
            raise ValueError("start_index can only be used if ordering is True")

        if ordering:
            sorted_halo_numbers = np.arange(start_index, start_index + len(halo_numbers_or_ordering))
            self.original_halo_numbers = sorted_halo_numbers[halo_numbers_or_ordering]
            self.sorted_to_unsorted_index = halo_numbers_or_ordering
            self.unsorted_to_sorted_index = np.argsort(halo_numbers_or_ordering)
        else:
            self.sorted_to_unsorted_index = np.argsort(halo_numbers_or_ordering)
            self.unsorted_to_sorted_index = np.argsort(self.sorted_to_unsorted_index)
            self.original_halo_numbers = halo_numbers_or_ordering
            sorted_halo_numbers = halo_numbers_or_ordering[self.sorted_to_unsorted_index]
        super().__init__(sorted_halo_numbers)

    def number_to_index(self, halo_number: Union[int, npt.NDArray[int]]) -> Union[int, npt.NDArray[int]]:
        halo_index_sorted = super().number_to_index(halo_number)
        return self.sorted_to_unsorted_index[halo_index_sorted]

    def index_to_number(self, halo_index: Union[int, npt.NDArray[int]]) -> Union[int, npt.NDArray[int]]:
        halo_index_sorted = self.unsorted_to_sorted_index[halo_index]
        return super().index_to_number(halo_index_sorted)

    def __iter__(self):
        yield from self.original_halo_numbers

    @property
    def all_numbers(self):
        """Returns the original halo numbers"""
        return self.original_halo_numbers


def create_halo_number_mapper(halo_numbers: npt.NDArray[int]) -> HaloNumberMapper:
    """Create the most efficient possible HaloNumberMapper for the given array of halo numbers"""
    halo_numbers = np.asarray(halo_numbers)
    if not np.issubdtype(halo_numbers.dtype, np.integer):
        raise ValueError("Halo number array must be integers")
    zero_offset = int(halo_numbers[0])

    # Check if halo_numbers can be represented by SimpleHaloNumberMapper
    if np.array_equal(halo_numbers, np.arange(zero_offset, len(halo_numbers) + zero_offset)):
        return SimpleHaloNumberMapper(zero_offset, len(halo_numbers))

    # Check if halo_numbers can be represented by MonotonicHaloNumberMapper
    if (np.diff(halo_numbers)>0).all():
        return MonotonicHaloNumberMapper(halo_numbers)

    # If neither of the above, use NonMonotonicHaloNumberMapper
    return NonMonotonicHaloNumberMapper(halo_numbers)



import numpy as np
import numpy.testing as npt
import pytest
from pytest import raises

import pynbody
import pynbody.halo.velociraptor
import pynbody.snapshot.swift


def test_load_identifies_swift():
    f = pynbody.load("testdata/SWIFT/snap_0150.hdf5")
    assert isinstance(f, pynbody.snapshot.swift.SwiftSnap)

def test_swift_properties():
    f = pynbody.load("testdata/SWIFT/snap_0150.hdf5")

    assert np.allclose(f.properties['a'], 0.38234515)
    assert np.allclose(f.properties['z'], 1.61543791)
    assert np.allclose(f.properties['h'], 0.703)
    assert np.allclose(f.properties['boxsize'].in_units("Mpc a h^-1", **f.conversion_context()), 100.)
    assert np.allclose(f.properties['omegaM0'], 0.276)
    assert np.allclose(f.properties['omegaL0'], 0.724)
    assert np.allclose(f.properties['omegaNu0'], 0.0)

def test_swift_arrays():
    f = pynbody.load("testdata/SWIFT/snap_0150.hdf5")
    assert np.allclose(f.dm['pos'].units.ratio("Mpc a", **f.conversion_context()), 1.0)
    assert np.allclose(f.dm['vel'].units.ratio("km s^-1", **f.conversion_context()), 1.0)
    # the reason the following isn't exactly 1.0 is because our solar mass is slightly different to swift's
    # (the pynbody value is outdated but it will need some work to think about how to fix this without
    # breaking backwards compatibility)
    assert np.allclose(f.dm['mass'].units.ratio("1e10 Msol", **f.conversion_context()), 1.0)
    assert np.allclose(f.dm['vel'][::50000], np.array([[-249.5395 ,   122.65865 , -144.79892 ],
                                             [  75.57313 ,  -51.598354 , 250.10258 ],
                                             [-139.62218 , -132.5298   , 479.02545 ],
                                             [ 147.22443 , -168.17662  ,-249.17387 ],
                                             [  27.643984,  161.06497  ,  21.430338],
                                             [  79.65777 ,   25.674492 , -45.813534]]))
    assert np.allclose(f.gas['pos'][::50000], np.array([[  2.54333146,   0.56501471,   3.08184457],
                                                        [ 27.33643304,  78.82288643,  62.55081956],
                                                        [ 50.74337741, 134.5097336 ,  95.28821923],
                                                        [ 86.53860045,  83.61342416, 129.95370508],
                                                        [111.7177823 ,  24.85736548,  55.55540164],
                                                        [128.76603044,  73.44601203, 139.65444299]]))

    assert np.allclose(float(f.dm['mass'].sum()+f.gas['mass'].sum()), 10895511.25)

def _assert_multifile_contents_is_sensible(f):
    assert len(f.dm) == 262144
    assert np.allclose(f.dm['pos'][::50000], np.array([[0.90574413, 1.23148826, 1.08457044],
                                                       [63.52396676, 5.42349734, 27.65864702],
                                                       [114.45492871, 9.92260142, 54.59853019],
                                                       [32.11092237, 32.28962438, 81.32341474],
                                                       [83.46050257, 20.91647755, 116.546989],
                                                       [125.83232028, 49.9732396, 72.2264199]]))

def test_swift_multifile_with_vds():
    f = pynbody.load("testdata/SWIFT/multifile_with_vds/snap_0000.hdf5")
    assert isinstance(f, pynbody.snapshot.swift.SwiftSnap)
    assert len(f._hdf_files) == 1
    assert f._hdf_files.is_virtual()
    _assert_multifile_contents_is_sensible(f)


def test_swift_multifile_without_vds():
    f = pynbody.load("testdata/SWIFT/multifile_without_vds/snap_0000")
    assert isinstance(f, pynbody.snapshot.swift.SwiftSnap)
    assert len(f._hdf_files) == 10
    assert not f._hdf_files.is_virtual()
    _assert_multifile_contents_is_sensible(f)

def test_swift_singlefile_is_not_vds():
    f = pynbody.load("testdata/SWIFT/snap_0150.hdf5")
    assert not f._hdf_files.is_virtual()

def test_swift_singlefile_partial_loading():
    f = pynbody.load("testdata/SWIFT/snap_0150.hdf5",
                     take_swift_cells=[5,15,20,25])
    assert len(f.dm) == 1849
    assert len(f.gas) == 1882
    assert (f.dm['iord'][::100] == [ 16468,   9172,  41176,  49874,   9342,  10234,  33908,  25852,
                                     42628,  34566,  26566,  10818, 502992,  67776,  34896,  68286,
                                     28052,  35988,  69524]).all()
    assert (f.gas['iord'][::100] == [16471,  8667, 49109, 57813, 17913, 26115, 10115, 50429, 50437,
                                     26367,  2503, 26825, 10827, 59463, 26703, 51909, 27927, 12057,
                                     36761]).all()

def test_swift_multifile_partial_loading():
    f = pynbody.load("testdata/SWIFT/multifile_without_vds/snap_0000",
                     take_swift_cells=[0,5,20,200])

    assert len(f) == 2048

    assert np.allclose(f['pos'][::100],
                       [[  0.90574413,   1.23148826,   1.08457044],
                          [  5.38928105,   3.39531932,  12.16627829],
                          [  0.94682741,   9.99268563,  14.42636433],
                          [ 16.65709673,   3.38461118,   5.47390269],
                          [  9.87028024,  14.62473673,   0.98727947],
                          [ 14.47500443,  16.773776  ,   9.74889636],
                          [ 54.2541276 ,  25.47568787,  14.10461116],
                          [ 58.85629125,  34.2701821 ,   7.44578035],
                          [ 67.83300647,  18.60943356,   0.78988041],
                          [ 63.32954201,  27.78617349,   0.87468019],
                          [ 67.95890776,  27.66693028,  16.39712484],
                          [  3.26246761,   1.20717194, 103.37377035],
                          [  5.47939785,  14.4774205 ,  92.18117726],
                          [  9.93556488,   5.69315011,  92.2609549 ],
                          [ 16.79507021,   7.98814361, 103.17866188],
                          [ 10.07660354,  16.88916246, 105.41116598],
                          [  1.3004053 ,  36.6884804 ,  81.22794421],
                          [  7.71374451,  45.49463307,  74.65383169],
                          [ 12.14706372,  38.8390906 ,  76.86117397],
                          [ 16.59894837,  36.67247821,  85.82433427],
                          [  9.79404938,  52.28051827,  81.30710868]])

def test_swift_multifile_partial_loading_order_insensitive():
    f = pynbody.load("testdata/SWIFT/multifile_without_vds/snap_0000",
                     take_swift_cells=[0, 5, 20, 200])
    f2 = pynbody.load("testdata/SWIFT/multifile_without_vds/snap_0000",
                     take_swift_cells=[0, 5, 20, 200][::-1])
    assert (f['iord'] == f2['iord']).all()

def test_swift_vds_partial_loading():
    f = pynbody.load("testdata/SWIFT/multifile_with_vds/snap_0000.hdf5", # <-- VDS file
                     take_swift_cells=[0,5,20,200])
    f2 = pynbody.load("testdata/SWIFT/multifile_without_vds/snap_0000",
                      take_swift_cells=[0, 5, 20, 200][::-1])
    assert (f['iord'] == f2['iord']).all()

def test_swift_fof_groups():
    f = pynbody.load("testdata/SWIFT/snap_0150.hdf5")
    h = f.halos(priority = ['HaloNumberCatalogue'])

    with raises(KeyError):
        _ = h[0]

    assert len(h[1]) == 444
    assert len(h[1].dm) == 247
    assert len(h[1].gas) == 197

    assert len(h) < 1000

def test_swift_dtypes():
    f = pynbody.load("testdata/SWIFT/snap_0150.hdf5")
    assert np.issubdtype(f['iord'].dtype, np.integer)
    assert np.issubdtype(f['pos'].dtype, np.floating)

@pytest.mark.parametrize('test_region',
                         [pynbody.filt.Sphere(50., (50., 50., 50.)),
                         pynbody.filt.Cuboid(-20.0)]) # note the cuboid test region wraps around the box
def test_swift_take_geometric_region(test_region):
    f = pynbody.load("testdata/SWIFT/snap_0150.hdf5",
                     take_region = test_region)

    f_full = pynbody.load("testdata/SWIFT/snap_0150.hdf5")

    assert len(f) < len(f_full)

    assert np.all(f[test_region]['iord'] == f_full[test_region]['iord'])

def test_swift_scalefactor_in_units():
    f = pynbody.load("testdata/SWIFT/snap_0150.hdf5")

    # naively, one would assume cm^2 s^-2, but actual units header says 1e10 a^2 cm^2 s^-2
    # So this tests pynbody respects that. Note that we don't give the scalefactor context,
    # so if scalefactor exponents are wrong, this will raise an exception
    npt.assert_allclose((f.gas['u'].units).in_units("km^2 s^-2 a^-2"), 1.0)

    npt.assert_allclose(f.gas['pos'].units.in_units("Mpc a"), 1.0)



import glob
import os.path
import pathlib
import shutil
import stat
import subprocess
import warnings

import numpy as np
import numpy.testing as npt
import pytest

import pynbody


@pytest.fixture
def cleanup_fpos_file():
    if os.path.exists("testdata/gasoline_ahf/g15784.lr.01024.AHF_fpos"):
        os.remove("testdata/gasoline_ahf/g15784.lr.01024.AHF_fpos")
    yield


def test_load_ahf_catalogue():
    f = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")
    h = pynbody.halo.ahf.AHFCatalogue(f)
    assert len(h)==1411

def test_load_ahf_catalogue_via_filename():
    f = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")
    f._filename = '' # obfuscate the filename, so we can check loading via user-specified filename

    h = f.halos(filename="testdata/gasoline_ahf/g15784.lr.01024.z0.000.AHF")
    assert isinstance(h, pynbody.halo.ahf.AHFCatalogue)
    assert len(h)==1411

h0_sample_iords = [57, 27875, 54094, 82969, 112002, 140143, 173567, 205840, 264606,
           301694, 333383, 358730, 374767, 402300, 430180, 456015, 479885, 496606,
           519824, 539971, 555195, 575204, 596047, 617669, 652724, 1533992, 1544021,
           1554045, 1564080, 1574107, 1584130, 1594158, 1604204, 1614257, 1624308, 1634376,
           1644485, 1654580, 1664698, 1674831, 1685054, 1695252, 1705513, 1715722, 1725900,
           1736070, 1746235, 1756400, 1766584, 1776754, 1786886]

@pytest.mark.parametrize("do_load_all", [True, False])
def test_ahf_particles(do_load_all, cleanup_fpos_file):
    f = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")
    h = pynbody.halo.ahf.AHFCatalogue(f)

    if do_load_all:
        h.load_all()

    assert len(h[0])==502300
    assert h[0].ancestor is f
    assert (h[0]['iord'][::10000]==h0_sample_iords).all()
    assert len(h[19])==3272
    assert(h[19]['iord'][::1000] == [232964, 341019, 752354, 793468]).all()

def test_load_copy():
    f = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")
    h = pynbody.halo.ahf.AHFCatalogue(f)
    hcopy = h.load_copy(0)
    assert (hcopy['iord'][::10000] == h0_sample_iords).all()
    assert hcopy.ancestor is not f

@pytest.mark.parametrize("do_load_all", [True, False])
def test_load_ahf_catalogue_non_gzipped(do_load_all):
    for extension in ["halos", "particles", "substructure"]:
        subprocess.call(["gunzip",f"testdata/gasoline_ahf/g15784.lr.01024.z0.000.AHF_{extension}.gz"])
    try:
        f = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")
        h = pynbody.halo.ahf.AHFCatalogue(f)
        if do_load_all:
            h.load_all()
        assert len(h)==1411
    finally:
        for extension in ["halos", "particles", "substructure"]:
            subprocess.call(["gzip", f"testdata/gasoline_ahf/g15784.lr.01024.z0.000.AHF_{extension}"])


def test_ahf_properties():
    f = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")
    h = pynbody.halo.ahf.AHFCatalogue(f)
    assert np.allclose(h[0].properties['Mvir'], 1.69639e+12)
    assert np.allclose(h[1].properties['Ekin'],6.4911e+17)
    assert np.allclose(h[1].properties['Mvir'], 1.19684e+13)

def test_ahf_all_properties():
    f = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")
    h = pynbody.halo.ahf.AHFCatalogue(f)

    # nb AHF reader currently doesn't infer units anyway, so with_units will make no difference
    properties = h.get_properties_all_halos(with_units=False)
    assert np.allclose(properties['Mvir'][0], 1.69639e+12)
    assert np.allclose(properties['Ekin'][1],6.4911e+17)
    assert np.allclose(properties['Mvir'][1], 1.19684e+13)

@pytest.fixture
def snap_in_unwritable_folder():
    if os.path.exists("testdata/test_unwritable"):
        os.chmod("testdata/test_unwritable", stat.S_IRUSR | stat.S_IXUSR | stat.S_IWUSR)
        shutil.rmtree("testdata/test_unwritable")
    os.mkdir("testdata/test_unwritable/")
    for fname in glob.glob("testdata/gasoline_ahf/*"):
        if "AHF_fpos" not in fname:
            os.symlink("../"+fname[9:], "testdata/test_unwritable/"+fname[22:])
    os.chmod("testdata/test_unwritable", stat.S_IRUSR | stat.S_IXUSR)

    yield "testdata/test_unwritable/g15784.lr.01024"

    os.chmod("testdata/test_unwritable", stat.S_IRUSR | stat.S_IXUSR | stat.S_IWUSR)
    shutil.rmtree("testdata/test_unwritable")



def test_ahf_unwritable(snap_in_unwritable_folder):
    f = pynbody.load(snap_in_unwritable_folder)

    # check we can still get a halo even without ability to write fpos file, but a warning is issued
    with pytest.warns(UserWarning, match="Unable to write AHF_fpos file;.*"):
        h = f.halos()
        _ = h[1]

    # check no subsequent warning issued
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        _ = h[2]

    # check if we use load_all there is no warning
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        h = f.halos()
        h.load_all()
        _ = h[1]


def test_detecting_ahf_catalogues_with_without_trailing_slash():
    # Test small fixes in #688 to detect AHF catalogues with and wihtout trailing slashes in directories
    for name in (
        "testdata/ramses_new_format_cosmo_with_ahf_output_00110",
        "testdata/ramses_new_format_cosmo_with_ahf_output_00110/"
    ):
        f = pynbody.load(name)
        _halos = pynbody.halo.ahf.AHFCatalogue(f)


def test_ramses_ahf_family_mapping_with_new_format():
    # Test Issue 691 where family mapping of AHF catalogues with Ramses new particle formats would go wrong
    f = pynbody.load("testdata/ramses_new_format_cosmo_with_ahf_output_00110")
    halos = pynbody.halo.ahf.AHFCatalogue(f)

    assert len(halos) == 149    # 150 lines in AHF halos file

    # Load halos and check that stars, DM and gas are correctly mapped by pynbody
    # Halo 0 is the main halo and has all three families, while other are random picks
    halo_numbers = [0, 9, 14]
    for halo_number in halo_numbers:
        halo = halos[halo_number]

        # There should not be any extra families in the halo particles
        assert(all(fam in [pynbody.family.dm, pynbody.family.star, pynbody.family.gas] for fam in halo.families()))

        # Check we now have the same number of particles assigned to the halo
        # than its AHF header, family by family
        assert halo.properties['npart'] == len(halo)
        assert halo.properties['n_star'] == len(halo.st)
        assert halo.properties['n_gas'] == len(halo.g)
        ndm = halo.properties['npart'] - halo.properties['n_star'] - halo.properties['n_gas']
        assert ndm == len(halo.d)

        # Derive some masses to check that we are identifying the right particles, in addition to their right numbers
        dm_mass = halo.properties['Mhalo'] - halo.properties['M_star'] - halo.properties['M_gas']
        gas_mass = halo.properties['M_gas']

        rtol = 1e-2 # We are not precise to per cent level with unit conversion through the different steps
        hubble = f.properties['h']      # AHF internal units are Msol/h and need to be manually corrected, which has not been done on this test output
        npt.assert_allclose(dm_mass / hubble, halo.d['mass'].sum().in_units("Msol"), rtol=rtol)
        npt.assert_allclose(gas_mass / hubble, halo.g['mass'].sum().in_units("Msol"), rtol=rtol)
        npt.assert_allclose(halo.properties['Mhalo'] / hubble, halo['mass'].sum().in_units("Msol"), rtol=rtol)

def test_ahf_corrupt_substructure():
    """For some reason lost in history, the AHF substructure file for gasoline_ahf/g15784 is corrupt.
    Use that to test we can see the exception"""
    f = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")
    with pytest.raises(KeyError):
        _ = pynbody.halo.ahf.AHFCatalogue(f, ignore_missing_substructure=False)


def test_ahf_substructure():
    f = pynbody.load("testdata/ramses_new_format_cosmo_with_ahf_output_00110")
    halos = pynbody.halo.ahf.AHFCatalogue(f)
    halos.load_all()

    check_parents = [0,1]
    check_children = [[72, 74, 138, 89], [115, 124,  20,  96]]

    for parent, children in zip(check_parents, check_children):
        halo = halos[parent]
        assert len(halo.properties['children']) == len(children)
        assert (halo.properties['children'] == children).all()
        assert halo.properties['hostHalo'] == -1
        #assert 'parent' not in halo.properties.keys()
        for child in children:
            assert halos[child].properties['parent'] == parent # from substructure file
            assert halos[child].properties['hostHalo'] == parent # from halos file

        assert len(halo.subhalos) == len(halo.properties['children'])
        for child_halo, child_number in zip(halo.subhalos, children):
            assert (child_halo.dm['iord'] == halos[child_number].dm['iord']).all()

@pytest.fixture
def snap_with_non_sequential_halos():
    base_folder = pathlib.Path("testdata/ahf_with_non_sequential_ids").absolute()
    if not base_folder.exists():
        base_folder.mkdir()
        for fname in base_folder.parent.glob("gasoline_ahf/*"):
            if "AHF" not in fname.name:
                (base_folder/fname.name).symlink_to(fname.absolute())

        # copy AHF_halos line by line, incorporating random IDs
        np.random.seed(0)
        my_random_ids = np.random.choice(1000000, 1411, replace=False)
        my_file_order = np.arange(1411)

        # shuffle the file ordering too
        np.random.shuffle(my_file_order)

        # my_file_order[:20] = np.arange(19,-1,-1) # reverse order of first 20

        with open(base_folder/"g15784.lr.01024.z0.000.AHF_halos", "w") as f:
            with pynbody.util.open_("testdata/gasoline_ahf/g15784.lr.01024.z0.000.AHF_halos", "rt") as f2:
                header = f2.readline()
                f.write("#ID(0) "+header)

                remaining_lines = f2.readlines()
                remaining_lines = [remaining_lines[i] for i in my_file_order]
                for line, id_ in zip(remaining_lines, my_random_ids[my_file_order]):
                    f.write(f"{id_} {line}")

        particle_file_per_halo = []
        with pynbody.util.open_("testdata/gasoline_ahf/g15784.lr.01024.z0.000.AHF_particles", "rt") as f2:
            assert f2.readline().strip() == "1411"
            for i in range(1411):
                lines = [f2.readline()]
                n_to_read = int(lines[0])
                for _ in range(n_to_read):
                    lines.append(f2.readline())
                particle_file_per_halo.append(lines)

        with open(base_folder/"g15784.lr.01024.z0.000.AHF_particles", "w") as f:
            f.write("1411\r\n")
            for o in my_file_order:
                f.writelines(particle_file_per_halo[o])

    return pynbody.load("testdata/ahf_with_non_sequential_ids/g15784.lr.01024")

@pytest.mark.parametrize("halo_numbering_mode, halo_ids",
                         [('ahf', (157105, 608171)),
                          ('v1', (761, 944)),
                          ('file-order', (760, 943)),
                          ('length-order-v1', (1,20)),
                          ('length-order', (0, 19))])
def test_ahf_non_sequential_ids(snap_with_non_sequential_halos,
                                halo_numbering_mode,
                                halo_ids):
    f = snap_with_non_sequential_halos
    h = pynbody.halo.ahf.AHFCatalogue(f, halo_numbers=halo_numbering_mode)

    assert len(h)==1411
    assert len(h[halo_ids[0]])==502300

    assert halo_ids[0] in h

    assert (h[halo_ids[0]]['iord'][::10000]==np.array([57, 27875, 54094, 82969, 112002, 140143, 173567, 205840, 264606,
           301694, 333383, 358730, 374767, 402300, 430180, 456015, 479885, 496606,
           519824, 539971, 555195, 575204, 596047, 617669, 652724, 1533992, 1544021,
           1554045, 1564080, 1574107, 1584130, 1594158, 1604204, 1614257, 1624308, 1634376,
           1644485, 1654580, 1664698, 1674831, 1685054, 1695252, 1705513, 1715722, 1725900,
           1736070, 1746235, 1756400, 1766584, 1776754, 1786886], dtype=np.int32)).all()

    assert len(h[halo_ids[1]])==3272
    assert(h[halo_ids[1]]['iord'][::1000] == [232964, 341019, 752354, 793468]).all()

def test_ahf_dosort_kwarg(snap_with_non_sequential_halos):
    # checks for compatibility with v1 behaviour
    f = snap_with_non_sequential_halos
    with pytest.warns(DeprecationWarning):
        h = pynbody.halo.ahf.AHFCatalogue(f, dosort=True)
    assert len(h)==1411
    assert len(h[1])==502300
    assert len(h[20])==3272



"""Support for the SubFind halo finder"""

from __future__ import annotations

import os.path
import warnings

import numpy as np

from .. import units
from ..array import SimArray
from . import HaloCatalogue
from .details import number_mapping, particle_indices
from .subhalo_catalogue import SubhaloCatalogue


class SubfindCatalogue(HaloCatalogue):
    """Handles catalogues produced by the SubFind halo finder (old versions that do not use HDF5 outputs)."""

    def __init__(self, sim, filename=None, subs=None, subhalos=False,  ordered=None, _inherit_data_from=None):
        """Initialise a SubFind catalogue

        By default, the FoF groups are imported, and subhalos are available via the 'subhalos' attribute of each
        halo object, e.g.

        >>> f = pynbody.load('path/to/snapshot')
        >>> h = f.halos()
        >>> h[1].subhalos[2] # returns the third subhalo of FoF group 1

        However by setting ``subs=True``, the FoF groups are ignored and the catalogue is of all subhalos.

        Parameters
        ----------

        sim : ~pynbody.snapshot.simsnap.SimSnap
            The simulation snapshot to which this catalogue applies.

        filename : str, optional
            The path to the SubFind output(s). This is expected to be a folder named ``groups_XXX`` where XXX is the
            snapshot number. The code extracts the snapshot number from the folder name and uses it to construct
            the filename of the catalogue files, for example ``groups_XXX/subhalo_tab_XXX.0``. If no filename is
            provided, the code will attempt to find a suitable catalogue starting from the simulation's directory.

        subhalos : bool, optional
            If False (default), catalogue represents the FoF groups and subhalos are available through the
            :meth:`~pynbody.halo.Halo.subhalos` attribute of each group (see note above). If True, the catalogue
            represents the subhalos directly and FoF groups are not available.

        ordered : bool, optional
            If True, the snapshot must have sequential iords. If False, the iords might be out-of-sequence, and
            therefore reordering will take place. If None (default), the iords are examined to check whether
            re-ordering is required or not. Note that re-ordering can be undesirable because it also destroys
            subfind's order-by-binding-energy

        _inherit_data_from : SubfindCatalogue, optional
            For internal use only; allows subhalo catalogue to share data with its parent FOF catalogue


        """

        if subs is not None:
            warnings.warn("The 'subs' argument to SubfindCatalogue is deprecated; use 'subhalos' instead",
                          DeprecationWarning)
            subhalos = subs

        self._subs=subhalos

        if ordered is not None:
            self._ordered = ordered
        else:
            self._ordered = bool((sim['iord']==np.arange(len(sim))).all())

        self._halos = {}

        self.dtype_int = sim['iord'].dtype
        self.dtype_flt = 'float32' #SUBFIND data is apparently always single precision???
        self._subfind_dir = filename or self._name_of_catalogue(sim)

        self.header = self._read_header()
        self._tasks = self.header[4]

        if _inherit_data_from:
            self._inherit_data(_inherit_data_from)
        else:
            self._read_data(sim)



        if subhalos:
            if self.header[6]==0:
                raise ValueError("This file does not contain subhalos")
            length = len(self._subhalodat['sub_off'])
        else:
            length = len(self._halodat['group_off'])

        super().__init__(sim, number_mapping.SimpleHaloNumberMapper(0, length))

        if not subhalos:
            self._subhalo_catalogue = SubfindCatalogue(sim, subhalos=True, filename=filename,
                                                       ordered=self._ordered, _inherit_data_from=self)

    def _inherit_data(self, from_):
        inherit = ['_halodat', '_subhalodat', '_keys_halo', '_keys_subhalo']
        for k in inherit:
            setattr(self, k, getattr(from_, k))

    def _get_all_particle_indices(self):
        ids = self._read_ids()
        boundaries = np.empty((len(self), 2), dtype=self.dtype_int)
        if self._subs:
            boundaries[:, 0] = self._subhalodat['sub_off']
            boundaries[:, 1] = self._subhalodat['sub_off'] + self._subhalodat['sub_len']
        else:
            boundaries[:, 0] = self._halodat['group_off']
            boundaries[:, 1] = self._halodat['group_off'] + self._halodat['group_len']

        if not self._ordered:
            self._init_iord_to_fpos()
            for a, b in boundaries:
                ids[a:b] = self._iord_to_fpos.map_ignoring_order(ids[a:b])
                # must be done segmented in case iord_to_fpos doesn't preserve input order

        return particle_indices.HaloParticleIndices(ids, boundaries)

    def get_properties_one_halo(self, i):

        extract = units.get_item_with_unit

        properties = {}

        if self._subs is False:
            for key in self._keys_halo:
                properties[key] = extract(self._halodat[key], i)
            properties['children'] = self._get_children_of_group(i)
        else:
            for key in self._keys_subhalo:
                properties[key] = extract(self._subhalodat[key], i)
            properties['children'] = self._get_children_of_sub(i)
        return properties

    def _get_children_of_group(self, group_nr):
        if self.header[6] > 0:
            return np.where(self._subhalodat['sub_groupNr'] == group_nr)[0]
        else:
            return []

    def _get_children_of_sub(self, subhalo_nr):
        # this goes down one level in the hierarchy, i.e. a subhalo will have all its sub-subhalos listed,
        # but not its sub-sub-subhalos (those will be listed in each sub-subhalo)
        halo_number_within_group = (self._subhalodat['sub_groupNr'][:subhalo_nr]
                                    == self._subhalodat['sub_groupNr'][subhalo_nr]).sum()
        return np.where((self._subhalodat['sub_parent'] == halo_number_within_group) &
                        (self._subhalodat['sub_groupNr'] == self._subhalodat['sub_groupNr'][subhalo_nr]))[0]


    def get_properties_all_halos(self, with_units=True) -> dict:
        properties = {}
        if self._subs:
            data = self._subhalodat
            properties['children'] = [self._get_children_of_sub(subhalo_nr)
                                      for subhalo_nr in self.number_mapper.all_numbers]
        else:
            data = self._halodat
            if self.header[6] > 0:
                properties['children'] = [self._get_children_of_group(group_nr)
                                          for group_nr in self.number_mapper.all_numbers]
            else:
                properties['children'] = []

        keys = self._keys_halo if not self._subs else self._keys_subhalo

        for key in keys:
            if with_units:
                properties[key] = data[key]
            else:
                properties[key] = data[key].view(np.ndarray)


        return properties

    def _read_header(self):
        iout = self._subfind_dir.split("_")[-1]
        filename = os.path.join(
            self._subfind_dir,
            f"subhalo_tab_{iout}.0"
        )
        with open(filename, "rb") as fd:
            # read header: this is strange but it works: there is an extra value in
            # header which we delete in the next step
            header = np.fromfile(fd, dtype='int32', sep="", count=8)

        header = np.delete(header, 4)

        return header

    def _get_subhalo_catalogue(self, parent_halo_number):
        if self._subs:
            return SubhaloCatalogue(self, self._get_children_of_sub(parent_halo_number))
        else:
            return SubhaloCatalogue(self._subhalo_catalogue,
                                    self._get_children_of_group(parent_halo_number))
    def _read_ids(self):
        data_ids = np.array([], dtype=self.dtype_int)
        iout = self._subfind_dir.split("_")[-1]
        for n in range(0, self._tasks):
            filename = os.path.join(
                self._subfind_dir,
                f"subhalo_ids_{iout}.{n}"
            )
            fd = open(filename, "rb")
            # for some reason there is an extra value in header which we delete
            # in the next step
            header1 = np.fromfile(fd, dtype='int32', sep="", count=7)
            header = np.delete(header1, 4)
            # TODO: include a check if both headers agree (they better)
            ids = np.fromfile(fd, dtype=self.dtype_int, sep="", count=-1)
            fd.close()
            data_ids = np.append(data_ids, ids)
        return data_ids

    def _read_data(self, sim):
        halodat={}
        keys_flt=['mass', 'pos', 'mmean_200', 'rmean_200', 'mcrit_200', 'rcrit_200', 'mtop_200', 'rtop_200', 'contmass']
        keys_int=['group_len', 'group_off',  'first_sub', 'Nsubs', 'cont_count', 'mostboundID']
        for key in keys_flt:
            halodat[key]=np.array([], dtype=self.dtype_flt)
        for key in keys_int:
            halodat[key]=np.array([], dtype='int32')

        subhalodat={}
        subkeys_int=['sub_len', 'sub_off', 'sub_parent', 'sub_mostboundID', 'sub_groupNr']
        subkeys_flt=['sub_pos', 'sub_vel', 'sub_CM', 'sub_mass', 'sub_spin', 'sub_veldisp', 'sub_VMax', 'sub_VMaxRad', 'sub_HalfMassRad', ]
        for key in subkeys_int:
            subhalodat[key]=np.array([], dtype='int32')
        subhalodat['sub_mostboundID']=np.array([], dtype=self.dtype_int)
        #subhalodat['sub_groupNr']=np.array([], dtype=self.dtype_int) #these are special
        for key in subkeys_flt:
            subhalodat[key]=np.array([], dtype=self.dtype_flt)

        self._keys_subhalo = subkeys_flt+subkeys_int
        self._keys_halo = keys_flt + keys_int

        for n in range(0,self._tasks):
            iout = self._subfind_dir.split("_")[-1]
            filename = os.path.join(
                self._subfind_dir,
                f"subhalo_tab_{iout}.{n}"
            )
            fd=open(filename, "rb")
            header1=np.fromfile(fd, dtype='int32', sep="", count=8)
            header=np.delete(header1,4)
            #read groups
            if header[0]>0:
                halodat['group_len']=np.append(halodat['group_len'], np.fromfile(fd, dtype='int32', sep="", count=header[0]))
                halodat['group_off']=np.append(halodat['group_off'], np.fromfile(fd, dtype='int32', sep="", count=header[0]))
                halodat['mass']=np.append(halodat['mass'], np.fromfile(fd, dtype=self.dtype_flt, sep="", count=header[0]))
                halodat['pos']=np.append(halodat['pos'], np.fromfile(fd, dtype=self.dtype_flt, sep="", count=3*header[0]) )
                halodat['mmean_200']=np.append(halodat['mmean_200'], np.fromfile(fd, dtype=self.dtype_flt, sep="", count=header[0]))
                halodat['rmean_200']=np.append(halodat['rmean_200'], np.fromfile(fd, dtype=self.dtype_flt, sep="", count=header[0]))
                halodat['mcrit_200']=np.append(halodat['mcrit_200'], np.fromfile(fd, dtype=self.dtype_flt, sep="", count=header[0]))
                halodat['rcrit_200']=np.append(halodat['rcrit_200'], np.fromfile(fd, dtype=self.dtype_flt, sep="", count=header[0]))
                halodat['mtop_200']=np.append(halodat['mtop_200'], np.fromfile(fd, dtype=self.dtype_flt, sep="", count=header[0]))
                halodat['rtop_200']=np.append(halodat['rtop_200'], np.fromfile(fd, dtype=self.dtype_flt, sep="", count=header[0]))
                halodat['cont_count']=np.append(halodat['cont_count'], np.fromfile(fd, dtype='int32', sep="", count=header[0]))
                halodat['contmass']=np.append(halodat['contmass'],np.fromfile(fd, dtype=self.dtype_flt, sep="", count=header[0]))
                halodat['Nsubs']=np.append(halodat['Nsubs'],np.fromfile(fd, dtype='int32', sep="", count=header[0]))
                halodat['first_sub']=np.append(halodat['first_sub'],np.fromfile(fd, dtype='int32', sep="", count=header[0]))
            #read subhalos only if expected to exist from header
            if header[5]>0:
                subhalodat['sub_len']=np.append(subhalodat['sub_len'], np.fromfile(fd, dtype='int32', sep="", count=header[5]))
                subhalodat['sub_off']=np.append(subhalodat['sub_off'], np.fromfile(fd, dtype='int32', sep="", count=header[5]))
                subhalodat['sub_parent']=np.append(subhalodat['sub_parent'], np.fromfile(fd, dtype='int32', sep="", count=header[5]))
                subhalodat['sub_mass']=np.append(subhalodat['sub_mass'], np.fromfile(fd, dtype=self.dtype_flt, sep="", count=header[5]))
                subhalodat['sub_pos']=np.append(subhalodat['sub_pos'],np.fromfile(fd, dtype=self.dtype_flt, sep="", count=3*header[5]))
                subhalodat['sub_vel']=np.append(subhalodat['sub_vel'], np.fromfile(fd, dtype=self.dtype_flt, sep="", count=3*header[5]))
                subhalodat['sub_CM']=np.append(subhalodat['sub_CM'], np.fromfile(fd, dtype=self.dtype_flt, sep="", count=3*header[5]))
                subhalodat['sub_spin']=np.append(subhalodat['sub_spin'], np.fromfile(fd, dtype=self.dtype_flt, sep="", count=3*header[5]))
                subhalodat['sub_veldisp']=np.append(subhalodat['sub_veldisp'], np.fromfile(fd, dtype=self.dtype_flt, sep="", count=header[5]))
                subhalodat['sub_VMax']=np.append(subhalodat['sub_VMax'], np.fromfile(fd, dtype=self.dtype_flt, sep="", count=header[5]))
                subhalodat['sub_VMaxRad']=np.append(subhalodat['sub_VMaxRad'],np.fromfile(fd, dtype=self.dtype_flt, sep="", count=header[5]))
                subhalodat['sub_HalfMassRad']=np.append(subhalodat['sub_HalfMassRad'], np.fromfile(fd, dtype=self.dtype_flt, sep="", count=header[5]))
                subhalodat['sub_mostboundID']=np.append(subhalodat['sub_mostboundID'], np.fromfile(fd, dtype=self.dtype_int, sep="", count=header[5]))
                subhalodat['sub_groupNr']=np.append(subhalodat['sub_groupNr'], np.fromfile(fd, dtype='int32', sep="", count=header[5]))
            fd.close()

        halodat['pos']=np.reshape(halodat['pos'], (header[1],3))
        if header[6]>0:
            #some voodoo because some SubFind files may have (at least?) one extra entry which is not really a subhalo
            real_ones=np.where(halodat['first_sub']<header[6])[0]
            fake_ones=np.where(halodat['first_sub']>=header[6])[0]
            halodat['mostboundID']=np.zeros(len(halodat['Nsubs']),dtype=self.dtype_int)-1
            halodat['mostboundID'][real_ones]=subhalodat['sub_mostboundID'][halodat['first_sub'][real_ones]]  #useful for the case of unordered snapshot IDs

            subhalodat['sub_pos']=np.reshape(subhalodat['sub_pos'], (header[6],3))
            subhalodat['sub_vel']=np.reshape(subhalodat['sub_vel'], (header[6],3))
            subhalodat['sub_CM']=np.reshape(subhalodat['sub_CM'], (header[6],3))
            subhalodat['sub_spin']=np.reshape(subhalodat['sub_spin'], (header[6],3))

        ar_names = 'mass', 'pos', 'mmean_200', 'rmean_200', 'mcrit_200', 'rcrit_200', 'mtop_200', 'rtop_200', \
                   'sub_mass', 'sub_pos', 'sub_vel', 'sub_CM', 'sub_veldisp', 'sub_VMax', 'sub_VMaxRad', 'sub_HalfMassRad'
        ar_dimensions = 'kg', 'm', 'kg', 'm', 'kg', 'm', 'kg', 'm', \
                    'kg', 'm', 'm s^-1', 'm', 'm s^-1', 'm s^-1', 'm', 'm'

        for name, dimension in zip(ar_names, ar_dimensions):
            if name in halodat:
                halodat[name] = SimArray(halodat[name], sim.infer_original_units(dimension))
                halodat[name].sim = sim
            if name in subhalodat:
                subhalodat[name] = SimArray(subhalodat[name], sim.infer_original_units(dimension))
                subhalodat[name].sim = sim

        self._halodat, self._subhalodat = halodat, subhalodat

    @staticmethod
    def _name_of_catalogue(sim):
        # standard path for multiple snapshot files
        snapnum = os.path.basename(
            os.path.dirname(sim.filename)).split("_")[-1]
        parent_dir = os.path.dirname(os.path.dirname(sim.filename))
        dir_path=os.path.join(parent_dir,"groups_" + snapnum)

        if os.path.exists(dir_path):
            return dir_path
        # alternative path if snapshot is single file
        else:
            snapnum = os.path.basename(sim.filename).split("_")[-1]
            parent_dir = os.path.dirname(sim.filename)
            return os.path.join(parent_dir,"groups_" + snapnum)

    @staticmethod
    def _can_load(sim, filename=None, **kwargs):
        if filename is not None:
            if str(filename)[:-3].endswith("groups_"):
                return True
        file = SubfindCatalogue._name_of_catalogue(sim)
        if os.path.exists(file):
            if os.path.exists(file):
                if file.endswith(".hdf5"):
                    return False
                elif os.path.isdir(file) and os.listdir(file)[0].endswith(".hdf5"):
                    return False
                else:
                    return True
        else:
            return False
